
import unittest

from metapack.test.test_basic import TestBasic
from metapack.test.test_packages import TestPackages
from metapack.test.test_ipy import TestIPython
from metapack.test.test_issues import TestIssues
from metapack.test.test_publish import TestPublish
from metapack.test.test_urls import TestUrls


def suite():

    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.makeSuite(TestBasic))
    test_suite.addTest(unittest.makeSuite(TestPackages))
    test_suite.addTest(unittest.makeSuite(TestIPython))
    test_suite.addTest(unittest.makeSuite(TestIssues))
    test_suite.addTest(unittest.makeSuite(TestPublish))
    test_suite.addTest(unittest.makeSuite(TestUrls))

    return test_suite


if __name__ == '__main__':
    mySuit=suite()
    runner=unittest.TextTestRunner()
    runner.run(mySuit)


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
CLI program for managing packages
"""

import re

from metapack import MetapackDoc, Downloader
from metapack.cli.core import prt, err, warn, metatab_info, write_doc, \
    update_name, process_schemas, extract_path_name, MetapackCliMemo
from metapack.util import make_metatab_file, datetime_now
from rowgenerators import parse_app_url
from rowgenerators.exceptions import SourceError
from rowgenerators.util import clean_cache
from tableintuit import RowIntuitError

downloader = Downloader()

def update(subparsers):
    parser = subparsers.add_parser(
        'update',
        help='Update a metatab file',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    parser.set_defaults(run_command=run_update)

    parser.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")



    parser.set_defaults(handler=None)

    ##
    ## Build Group

    group = parser.add_mutually_exclusive_group()

    group.add_argument('-s', '--schemas', default=False, action='store_true',
                             help='Rebuild the schemas for files referenced in the resource section')

    group.add_argument('-n', '--name', action='store_true', default=False,
                             help="Update the Name from the Datasetname, Origin and Version terms")

    parser.add_argument('-C', '--clean', action='store_true', default=False,
                        help='When building a schema, rebuild completely')

    parser.add_argument('-F', '--force', action='store_true', default=False,
                             help='Force the operation')


def run_update(args):

    m = MetapackCliMemo(args, downloader)


    if m.args.schemas:
        update_name(m.mt_file, fail_on_missing=False, report_unchanged=False)

        process_schemas(m.mt_file, cache=m.cache, clean=m.args.clean)

    if m.mtfile_url.scheme == 'file' and m.args.name:
        update_name(m.mt_file, fail_on_missing=True, force=m.args.force)



import unittest

from metapack import MetapackDoc
from metapack import MetapackUrl, Downloader
from metapack.terms import Reference, Resource
from metapack.test.support import test_data
from metatab import TermParser
from metatab.generate import TextRowGenerator


class TestBasic(unittest.TestCase):

    def test_resolve_packages(self):


        def u(v):
            return "http://example.com/d/{}".format(v)

        def f(v):
            return "file:/d/{}".format(v)

        for us in (u('package.zip'), u('package.xlsx'), u('package.csv'), u('package/metadata.csv'),
                   f('package.zip'), f('package.xlsx'), f('package.csv'), f('package/metadata.csv'),):

            u = MetapackUrl(us, downloader=Downloader())

            print(u.metadata_url)


    def test_open_package(self):

        from metapack import open_package
        from metapack.terms import Resource

        p = open_package(test_data('packages/example.com/example.com-full-2017-us/metadata.csv'))

        self.assertEqual(Resource, type(p.find_first('root.datafile')))

        self.assertEqual('example.com-full-2017-us-1', p.find_first('Root.Name').value)

        self.assertEqual(16, len(list(p['Resources'].find('Root.Resource'))))


        all_names = [ r.name for r in p.find('Datafile') ]
        for name in ['renter_cost', 'simple-example-altnames', 'simple-example', 'unicode-latin1', 'unicode-utf8',
                     'renter_cost_excel07', 'renter_cost_excel97', 'renter_cost-2', 'random-names',
                     'random-names-fs', 'random-names-csv', 'random-names-xlsx', 'random-names-zip', 'sra']:
            self.assertIn(name, all_names)



        self.assertIsInstance (p.resource('random-names'), Resource)
        self.assertEqual('random-names', p.resource('random-names').name)

        r = p.find_first('Root.DataFile')
        print(r.resolved_url)
        self.assertEqual('http://public.source.civicknowledge.com/example.com/sources/test_data.zip#renter_cost.csv',
                          str(r.resolved_url))

        for r in p.find('Root.DataFile'):

            if r.name != 'unicode-latin1':
                continue

            self.assertEqual(int(r.nrows), len(list(r)))

        self.assertEqual(['ipums', 'bordley', 'mcdonald', 'majumder'],
                          [c.name for c in p['Bibliography']])

    def test_line_oriented(self):

        doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest"))

        with open(test_data('line','line-oriented-doc.txt')) as f:
            text = f.read()

        tp = TermParser(TextRowGenerator(text), resolver=doc.resolver, doc=doc)

        doc.load_terms(tp)

        self.assertEqual('47bc1089-7584-41f0-b804-602ec42f1249', doc.get_value('Root.Identifier'))
        self.assertEqual(148, len(doc.terms))

        self.assertEqual(5, len(list(doc['References'])))

        self.assertEqual(5, len(list(doc['References'].find('Root.Reference'))))

        self.assertEqual(5, len(list(doc['References'].find('Root.Resource'))))  # References are Resources

        rt = list(doc['References'].find('Root.Resource'))[0]

        self.assertIsInstance(rt, Reference)

    def test_line_doc_parts(self):

        doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest"))

        for fn in ('line/line-oriented-doc-root.txt',
                   'line/line-oriented-doc-contacts.txt',
                   'line/line-oriented-doc-datafiles.txt',
                   'line/line-oriented-doc-references-1.txt',
                   'line/line-oriented-doc-references-2.txt',
                   'line/line-oriented-doc-bib.txt',
                   ):
            with open(test_data(fn)) as f:
                text = f.read()

            tp = TermParser(TextRowGenerator(text), resolver=doc.resolver, doc=doc)

            doc.load_terms(tp)

        self.assertEqual('47bc1089-7584-41f0-b804-602ec42f1249', doc.get_value('Root.Identifier'))
        self.assertEqual(153, len(doc.terms))

        self.assertEqual(5, len(list(doc['References'])))

        self.assertEqual(5, len(list(doc['References'].find('Root.Reference'))))

        self.assertEqual(5, len(list(doc['References'].find('Root.Resource'))))  # References are Resources

        rt = list(doc['References'].find('Root.Resource'))[0]

        self.assertIsInstance(rt, Reference)

        self.assertEqual(5, len(list(doc['Resources'])))

        self.assertEqual(5, len(list(doc['Resources'].find('Root.Datafile'))))

        self.assertEqual(5, len(list(doc['Resources'].find('Root.Resource'))))  # References are Resources

        rt = list(doc['Resources'].find('Root.Resource'))[0]

        self.assertIsInstance(rt, Resource)

        doc._repr_html_() # Check no exceptions



if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""

from metapack.cli.core import prt, err, warn
from metapack.package import *
from .core import MetapackCliMemo as _MetapackCliMemo
from tabulate import tabulate
import sys

downloader = Downloader()

class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)
        self


def doc_args(subparsers):

    parser = subparsers.add_parser(
        'doc',
        help='Generate documentation'
    )

    cmdsp = parser.add_subparsers(help='sub-command help')

    cmdp = cmdsp.add_parser('graph', help='Dependency graph')
    cmdp.set_defaults(run_command=graph_cmd)

    cmdp.add_argument('-d', '--dependencies', default=False, action='store_true',
                        help="List the dependencies")

    cmdp.add_argument('-v', '--view', default=False, action='store_true',
                       help="View doc file after creating it")

    cmdp.add_argument('-V', '--graphviz', default=False, action='store_true',
                       help="Also dump the Graphviz file to the stdout")

    cmdp.add_argument('-D', '--directory',  help="Output file name")

    cmdp.add_argument('-N', '--nonversion', default=False, action='store_true',
                      help="Use the nonversioned package name for the file name")

    cmdp.add_argument('metatabfile', nargs='?',
                      help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    ##
    ##
    cmdp = cmdsp.add_parser('deps', help='Display a table of dependencies')
    cmdp.set_defaults(run_command=deps_cmd)

    cmdp.add_argument('-p', '--packages', default=False, action='store_true',
                        help="When listing dependencies, list only packages")

    cmdp.add_argument('-f', '--format', default='pipe',
                      help="Table format, an argument to tabulate(). Common options are:"
                      " plain, simple, jira, html, pipe, rst, mediawiki. Use pipe for markdown.")

    cmdp.add_argument('metatabfile', nargs='?',
                      help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    ##
    ##
    cmdp = cmdsp.add_parser('schema', help='Print a schema')
    cmdp.set_defaults(run_command=dump_schemas)

    cmdp.add_argument('-f', '--format', default='pipe',
                      help="Table format, an argument to tabulate(). Common options are:"
                      " plain, simple, jira, html, pipe, rst, mediawiki. Use pipe for markdown.")

    cmdp.add_argument('-c', '--column', action='append',
                      help="Add a column from the schema to the output table. If specified, only the"
                           "'name' column is included by default")

    cmdp.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

def graph_cmd(args):

    m = MetapackCliMemo(args, downloader)

    graph(m)

from functools import total_ordering

@total_ordering
class _DependencyNode(object):

    @property
    def name(self):
        return self.node['name']

    @property
    def label(self):
        return self.node['label']

    @property
    def shape(self):
        return self.node['shape']

    def slugify(self,v):
        from rowgenerators.util import slugify

        return slugify(v)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    def __str__(self):
        return '<{} {} {} {}>'.format(type(self).__name__, self.name, self.label, self.shape)


class DependencyNode(_DependencyNode):

    def __new__(cls, ref, parent):
        from metapack import MetapackDoc
        from metapack.appurl import is_metapack_url
        from rowgenerators.appurl.sql import Sql

        if isinstance(ref,MetapackDoc):
            return PackageDependencyNode(ref, parent)

        u = ref.expanded_url

        if is_metapack_url(u):
            return PackageDependencyNode(u.doc, parent)
        elif isinstance(u,Sql):
            return SqlDependencyNode(ref, parent)
        elif u:
            return UrlDependencyNode(u, parent)
        else:
            return ResourceDependencyNode(ref, parent)


class PackageDependencyNode(_DependencyNode):
    """Dependency node that references a metapack package"""
    def __init__(self, ref, parent):
        self.doc = ref
        self.parent = parent

    @property
    def node(self):
        return {
            'name': self.slugify(self.doc.name),
            'label': self.doc.name,
            'shape': 'component'
        }

class UrlDependencyNode(_DependencyNode):
    """Dependency node that references a url"""
    def __init__(self, ref, parent):
        self.url = ref
        self.parent = parent

    def _shape(self):
        if self.url.scheme == 'file':
            return 'folder'
        elif self.url.scheme.startswith('http'):
            if self.url.proto != self.url.scheme:
                return 'cds' # An HTTP based API
            else:
                return 'note'
        else:
            return 'box'

    def _label(self):
        if self.url.scheme == 'file':
            return self.url.path
        elif self.url.scheme.startswith('http'):
            if self.url.scheme != self.url.proto:
                proto = '{}+{}'.format(self.url.scheme,self.url.proto)
            else:
                proto = '{}'.format(self.url.scheme)

            return "Proto: {}\nHost: {}\nPath: {}\nApi: {}".format(proto, self.url.netloc,self.url.path,
                                                                 self.url.fragment[0])
        else:
            return str(self.url)

    @property
    def node(self):
        return {
            'name': self.slugify(str(self.url)),
            'label': str(self.url),
            'shape': self._shape()
        }

class ResourceDependencyNode(_DependencyNode):
    """Dependency node that references a resoruce where the URL doesn't expand"""
    def __init__(self, ref, parent):
        self.resource = ref
        self.parent = parent

    @property
    def node(self):
        return {
            'name': self.slugify(self.resource.name),
            'label': self.resource.name,
            'shape': 'oval'
        }

class SqlDependencyNode(_DependencyNode):

    def __init__(self, ref, parent):
        self.resource = ref
        self.parent = parent

    @property
    def node(self):
        return {
            'name': self.slugify(self.resource.name),
            'label': self.resource.name,
            'shape': 'cylinder'
        }

def yield_deps(doc):
    """
    TODO: This implementation is very specific to a particular environment and project, and should be generalized.
    :param doc:
    :return:
    """

    this_node = DependencyNode(doc, None)

    for r in doc.references():

        other_node = DependencyNode(r, this_node)

        yield (this_node, other_node)

        if isinstance(other_node, PackageDependencyNode):
            yield from yield_deps(other_node.doc)

def nodes_edges(m):
    nodes = set()
    edges = set(yield_deps(m.doc))

    for this, that in edges:
        nodes.add(this)
        nodes.add(that)

    return nodes, list(edges)

def deps_cmd(args):

    m = MetapackCliMemo(args, downloader)

    nodes, edges = nodes_edges(m)

    if m.args.packages:
        for n in nodes:
            if isinstance(n,PackageDependencyNode):
                print(n.label)

    else:
        rows = [ (this.label, that.label) for this, that in sorted(edges) ]

        print (tabulate(rows,headers='This That'.split()))


def wrap_url(s, l):
    """Wrap a URL string"""
    parts = s.split('/')

    if len(parts) == 1:
        return parts[0]
    else:
        i = 0
        lines = []
        for j in range(i, len(parts) + 1):
            tv = '/'.join(parts[i:j])
            nv = '/'.join(parts[i:j + 1])

            if len(nv) > l or nv == tv:
                i = j
                lines.append(tv)

        return '/\n'.join(lines)



def legend_subgraph(g):

    with g.subgraph(name='cluster1') as c:
        c.attr(rank='sink')
        #c.attr(rankdir='TB')
        c.attr(labeltoc='t')
        c.attr(label='Legend')
        c.node(name='Datapackage', label='datapackage', shape='component')
        c.node(name='Database', label='Database', shape='cylinder')
        c.node(name='DataFile', label='Data File', shape='folder')
        c.node(name='Other', label='Other Datafile', shape='oval')
        c.node(name='Rest', label='REST', shape='cds')
        c.node(name='Web', label='Web Doc', shape='note')

        c.edge('Datapackage', 'Database', style='invis')
        c.edge('Database','DataFile', style='invis')
        #c.edge('DataFile', 'Other')
        c.edge('Other', 'Rest', style='invis')
        c.edge('Rest', 'Web', style='invis')

def graph(m):
    """

    :param m:
    :return:
    """
    from graphviz import Digraph

    from os.path import splitext, dirname

    if m.args.nonversion:
        output = m.doc.nonver_name + '.jpg'
    else:
        output = m.doc.name + '.jpg'


    basename, format = splitext(output)

    g = Digraph(comment=m.doc.name, filename=basename+'.gv', directory=m.args.directory)

    legend_subgraph(g)

    nodes, edges = nodes_edges(m)

    for n in nodes:
        g.node(name=n.name, label=wrap_url(n.label,40), shape = n.shape)

    for this, that in edges:
        g.edge(that.name, this.name)

    g.attr(overlap='false')
    g.attr(rankdir='LR')

    g.format = format.lstrip('.')
    g.engine = 'dot'
    g.render(basename, view=m.args.view, cleanup=True)

    if m.args.graphviz:
        prt(g.source)

    if not m.args.view:
        prt("Wrote file "+output)




def dump_schemas(args):
    from .run import get_resource

    m = MetapackCliMemo(args, downloader)

    r = m.doc.resource(m.resource)

    if not r:
        # Try a reference, which must be resolved to a resource in the foreign data package
        # to get a schema.
        outer_r = m.doc.reference(m.resource)

        if outer_r:
            try:
                r = outer_r.expanded_url.resource
            except AttributeError:
                warn("Reference '{}' is not a Metapack url, doesn't have a schema".format(m.resource))
        else:
            r = None


    if not r:
        prt('\nSelect a resource or reference to display the schema for:\n')
        d = []
        for r in m.doc.resources():
            d.append(('Resource', r.name, r.url))

        for r in m.doc.references():
            d.append(('Reference', r.name, r.url))


        prt(tabulate(d, 'Type Name Url'.split()))

        prt('')
        prt("Specify the resource as a fragment, escaping it if the '#' is the first character. For instance: ")
        prt("  mp doc schema .#resource_name")
        prt('')
        sys.exit(0)

    dump_schema(m, r)

def dump_schema(m, r):

    st = r.schema_term
    rows_about_columns = []

    if m.args.column:

        headers = ['Name']+[e.title() for e in m.args.column]

        for c in st.find('Table.Column'):
            table_row = [c.name]
            for table_col in m.args.column:
                table_row.append(c.get_value(table_col))
            rows_about_columns.append(table_row)

    else:
        headers = 'Name AltName DataType Description'.split()
        for c in st.find('Table.Column'):
            rows_about_columns.append((c.name, c.get_value('altname'), c.get_value('datatype'), c.get_value('description')))

    prt(tabulate(rows_about_columns, headers=headers, tablefmt=m.args.format))



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
NBConvert preprocessors
"""

from textwrap import dedent

from IPython.core.magic_arguments import (parse_argstring)
from nbconvert.preprocessors import Preprocessor
from nbformat.notebooknode import from_dict
from traitlets import Unicode, List, Dict

from metapack import MetapackDoc
from metatab import TermParser
from metatab.generate import TextRowGenerator
from rowgenerators import parse_app_url
from .magic import MetatabMagic


class RemoveDocsFromImages(Preprocessor):
    """Change the file name for images, because they are in the sam dir as the HTML files"""
    doc = None

    def preprocess_cell(self, cell, resources, index):

        for o in cell.get('outputs', {}):

            if not 'metadata' in o:
                continue

            image_file = o.get('metadata', {}).get('filenames', {}).get('image/png')

            if image_file:
                o['metadata']['filenames']['image/png'] = image_file.replace('docs/', '')

        return cell, resources


class ScrubPlainText(Preprocessor):
    """If there is a Latex or HTML representation for data, remove only latex representations
    This is primarily for bibliographies, where the text/plain version of the bib, which is
    just the repr() of an HTML object, is showing up in LaTex
    """
    doc = None

    def preprocess_cell(self, cell, resources, index):

        for o in cell.get('outputs', {}):

            d = o.get('data', {})

            if ('text/latex' in d or 'text/html' in d) and 'text/plain' in d:
                del d['text/plain']

        return cell, resources


class HtmlBib(Preprocessor):
    """ Keep only HTML outputs  for biblographies
    """

    remove_type = 'text/html'

    def preprocess_cell(self, cell, resources, index):

        if 'mt_bibliography' in cell.source or 'mt_data_references' in cell.source:
            outputs = []
            for o in cell.get('outputs', {}):

                d = o.get('data', {})
                remove_keys = [k for k in d.keys() if k != self.remove_type]
                o['data'] = {k: v for k, v in d.items() if k not in remove_keys}

                if o['data']:
                    outputs.append(o)

            cell['outputs'] = outputs

        return cell, resources


class LatexBib(HtmlBib):
    """ Keep only Latex outputs  for biblographies
    """

    remove_type = 'text/latex'


class MoveTitleDescription(Preprocessor):
    """Look for Markdown cells with Title or Description tags, clear them out. The ExtractMetatabTerms
    preprocessor is used to move these values into the Metatab metadata"""

    def preprocess(self, nb, resources):

        r = super().preprocess(nb, resources)

        return r

    def preprocess_cell(self, cell, resources, index):

        if cell['cell_type'] == 'markdown':
            tags = cell['metadata'].get('tags', [])

            if 'Title' in tags:
                m = resources.get('metadata', {})
                m['name'] = cell.source.strip().replace('#', '')
                resources['metadata'] = m
                cell.source = ''

            if 'Description' in tags:
                cell.source = ''

        return cell, resources


class ExtractMetatabTerms(Preprocessor):
    """Look for tagged markdown cells and use the value to set some metatab doc terms"""

    terms = None

    def preprocess_cell(self, cell, resources, index):

        if not self.terms:
            self.terms = []

        if cell['cell_type'] == 'markdown':

            tags = cell['metadata'].get('tags', [])

            if 'Title' in tags:
                self.terms.append(('Root', 'Root.Title', cell.source.strip().replace('#', '')))

            elif 'Description' in tags:
                self.terms.append(('Root', 'Root.Description', cell.source.strip()))

        return cell, resources


class ExtractInlineMetatabDoc(ExtractMetatabTerms):
    """Extract the Inlined Metatab document. Will Apply the metatab cell vaules for
    the Title and Description to the document terms. """

    package_url = Unicode(help='Metapack Package Url').tag(config=True)

    doc = None

    extra_terms = None

    def preprocess_cell(self, cell, resources, index):
        import re
        from metatab.generate import TextRowGenerator

        if not self.extra_terms:
            self.extra_terms = []

        if cell['source'].startswith('%%metatab'):

            tp = TermParser(TextRowGenerator(re.sub(r'\%\%metatab.*\n', '', cell['source'])),
                            resolver=self.doc.resolver, doc=self.doc)

            self.doc.load_terms(tp)

        elif cell['cell_type'] == 'markdown':

            tags = cell['metadata'].get('tags', [])

            if 'Title' in tags:
                self.extra_terms.append(('Root', 'Root.Title', cell.source.strip().replace('#', '')))

            elif 'Description' in tags:
                self.extra_terms.append(('Root', 'Root.Description', cell.source.strip()))

        else:
            cell, resources = super().preprocess_cell(cell, resources, index)

        return cell, resources

    def preprocess(self, nb, resources):

        r = super().preprocess(nb, resources)

        for section, term, value in self.terms:
            self.doc[section].get_or_new_term(term, value)

        return r

    def run(self, nb):

        assert str(self.package_url)

        self.doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest\n"),
                               package_url=parse_app_url(self.package_url))

        self.preprocess(nb, {})

        for section, term, value in self.extra_terms:
            self.doc[section].get_or_new_term(term, value)

        return self.doc


class ExtractFinalMetatabDoc(Preprocessor):
    """Extract the metatab document produced from the %mt_show_metatab magic"""

    doc = None

    def preprocess_cell(self, cell, resources, index):
        from metatab.generate import TextRowGenerator

        if cell['metadata'].get('mt_final_metatab'):
            if cell['outputs']:
                o = ''.join(e['text'] for e in cell['outputs'])

                self.doc = MetapackDoc(TextRowGenerator(o))

                # Give all of the sections their standard args, to make the CSV versions of the doc
                # prettier

                for name, s in self.doc.sections.items():
                    try:
                        s.args = self.doc.decl_sections[name.lower()]['args']
                    except KeyError:
                        pass

        return cell, resources


class ExtractMaterializedRefs(Preprocessor):
    """Extract the metatab document produced from the %mt_show_metatab magic"""

    materialized = None

    def preprocess_cell(self, cell, resources, index):

        from json import loads

        if cell['metadata'].get('mt_materialize'):

            if cell['outputs']:
                o = ''.join(e['text'] for e in cell['outputs'])

                self.materilized = loads(o)

        return cell, resources


class ExtractLibDirs(Preprocessor):
    """Extract the metatab document produced from the %mt_show_metatab magic"""

    lib_dirs = []

    def preprocess_cell(self, cell, resources, index):

        from json import loads

        if cell['metadata'].get('mt_show_libdirs'):

            if cell['outputs']:
                o = ''.join(e['text'] for e in cell['outputs'])

                if o:
                    self.lib_dirs = loads(o)

        return cell, resources


class RemoveMetatab(Preprocessor):
    """NBConvert preprocessor to remove the %metatab block"""

    def preprocess(self, nb, resources):

        out_cells = []

        for cell in nb.cells:

            source = cell['source']

            if cell['metadata'].get('epilog'):
                continue

            if source.startswith('%%metatab'):
                lines = source.splitlines()  # resplit to remove leading blank lines

                args = parse_argstring(MetatabMagic.metatab, lines[0].replace('%%metatab', ''))

                cell.source = "%mt_open_package\n"
                cell.outputs = []

                cell['metadata']['hide_input'] = True
                cell['metadata']['show_input'] = 'hide'
                cell['metadata']['hide_output'] = True

            out_cells.append(cell)

        nb.cells = out_cells

        return nb, resources


class RemoveMagics(Preprocessor):
    """Remove line magic lines, or entire cell magic cells"""

    def preprocess(self, nb, resources):
        import re

        for i, cell in enumerate(nb.cells):

            if re.match(r'^\%\%', cell.source):
                cell.source = ''
            else:
                cell.source = re.sub(r'\%[^\n]+\n?', '', cell.source)

        return nb, resources


class PrepareScript(Preprocessor):
    """Add an import so converted scripts can handle some magics"""

    def preprocess(self, nb, resources):
        nb.cells = [from_dict({
            'cell_type': 'code',
            'outputs': [],
            'metadata': {},
            'execution_count': None,
            'source': dedent("""
            from metatab.jupyter.script import get_ipython
            """)
        })] + nb.cells
        return nb, resources


class ReplaceMagics(Preprocessor):
    """Replace some magics"""


class NoShowInput(Preprocessor):
    """NBConvert preprocessor to add hide_input metatab to cells, except to cells that have either
     an %mt_showinput magic, or a 'show' tag """

    def preprocess(self, nb, resources):
        import re

        out_cells = []

        for cell in nb.cells:

            #  Code cells aren't displayed at all, unless it starts with
            # a '%mt_showinput' magic, which is removed

            if cell['cell_type'] == 'code':

                source = cell['source']

                tags = cell['metadata'].get('tags', [])

                if source.startswith('%mt_showinput') or 'show' in tags:
                    cell['source'] = re.sub(r'\%mt_showinput', '', source)
                else:
                    cell['metadata']['hide_input'] = True
                    cell['metadata']['show_input'] = 'hide'

            out_cells.append(cell)

        nb.cells = out_cells

        return nb, resources


class AddEpilog(Preprocessor):
    """Add final cells that writes the Metatab file, materializes datasets, etc.  """

    pkg_dir = Unicode(help='Metatab package Directory').tag(config=True)

    dataframes = List(help='Names of dataframes to materialize', trait=Unicode())

    def preprocess(self, nb, resources):
        from datetime import datetime

        # Well, now we adding prolog in the epilog ...
        nb.cells = [
                       from_dict({
                           'cell_type': 'code',
                           'outputs': [],
                           'metadata': {'': True, 'prolog': True},
                           'execution_count': None,
                           'source': ("#{}\n".format(datetime.now())) +
                                     "%load_ext metapack.jupyter.magic"
                       })
                   ] + nb.cells

        assert self.pkg_dir

        if len(self.dataframes) > 0:
            nb.cells.append(from_dict({
                'cell_type': 'code',
                'outputs': [],
                'metadata': {'mt_dataframes': True, 'epilog': True},
                'execution_count': None,
                'source': '\n'.join("%mt_materialize {} '{}' ".format(df, self.pkg_dir) for df in self.dataframes)
            }))

        nb.cells.append(from_dict({
            'cell_type': 'code',
            'outputs': [],
            'metadata': {'mt_materialize': True, 'epilog': True},
            'execution_count': None,
            'source': dedent("""
            %mt_materialize_all '{pkg_dir}'
            """.format(pkg_dir=self.pkg_dir))
        }))

        nb.cells.append(from_dict({
            'cell_type': 'code',
            'outputs': [],
            'metadata': {'mt_final_metatab': True, 'epilog': True},
            'execution_count': None,
            'source': dedent("""
            %mt_show_metatab

            """.format(pkg_dir=self.pkg_dir))
        }))

        nb.cells.append(from_dict({
            'cell_type': 'code',
            'outputs': [],
            'metadata': {'mt_show_libdirs': True, 'epilog': True},
            'execution_count': None,
            'source': dedent("""
            %mt_show_libdirs

            """.format(pkg_dir=self.pkg_dir))
        }))

        return nb, resources

class AddProlog(Preprocessor):
    """Add final cells that writes the Metatab file, materializes datasets, etc.  """

    env = Dict(help='Initial local variables. Must be primitive types').tag(config=True)

    def preprocess(self, nb, resources):
        from datetime import datetime

        source = '\n'.join( '{}={}'.format(k,repr(v)) for k,v in self.env.items()
                            if k and isinstance(v,(int,float,str)))

        nb.cells = [
                       from_dict({
                           'cell_type': 'code',
                           'outputs': [],
                           'metadata': {'': True, 'prolog': True},
                           'execution_count': None,
                           'source': source
                       }),
                    # Mark this notebook as being run in a build, which can change the behaviour of some magics
                       from_dict({
                           'cell_type': 'code',
                           'outputs': [],
                           'metadata': {'': True, 'prolog': True},
                           'execution_count': None,
                           'source': "METAPACK_BUILDING=True"
                       }),

                   ] + nb.cells

        return nb, resources



class OrganizeMetadata(Preprocessor):
    """Move a lot of metadata around """

    package_url = Unicode(help='Metapack Package Url').tag(config=True)

    alt_title = None  # Title extracted from markdown, without a Title tag

    show_input = 'show'

    def __init__(self, **kw):
        self.front_matter = {'show_input': self.show_input}
        self.metadata = {}

        self.doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest\n"),
                               package_url=parse_app_url(self.package_url))

        super().__init__(**kw)

    def preprocess(self, nb, resources):
        from metapack.util import slugify
        import uuid

        r = super().preprocess(nb, resources)

        nb.cells = [cell for cell in nb.cells if cell.source]

        self.front_matter['slug'] = slugify(self.front_matter.get('title', self.alt_title or str(uuid.uuid4())))

        # move the metadata we collected back into the notebook metadata.
        nb.metadata['frontmatter'] = self.front_matter
        nb.metadata['metatab'] = self.doc.as_dict()
        nb.metadata['metapack'] = self.metadata

        # Reset the identity, name and descriptions, because they can have add forms
        # in the data, which the Doc interface fixes.
        nb.metadata['metatab']['name'] = self.doc.name
        nb.metadata['metatab']['description'] = self.doc.description
        nb.metadata['metatab']['identifier'] = self.doc.identifier

        if not 'identifier' in nb.metadata['frontmatter']:
            nb.metadata['frontmatter']['identifier'] = nb.metadata['metatab']['identifier']

        return nb, resources

    def preprocess_cell(self, cell, resources, index):
        import yaml
        import re

        tags = [e.lower() for e in cell['metadata'].get('tags', [])]

        if 'frontmatter' in tags:
            d = yaml.load(cell['source'])
            self.front_matter.update(d)
            cell.source = ''

        if 'metadata' in tags:
            d = yaml.load(cell['source'])
            self.metadata.update(d)
            cell.source = ''

        if 'title' in tags:
            self.front_matter['title'] = cell.source.strip().replace('#', '')
            m = resources.get('metadata', {})
            m['name'] = cell.source.strip().replace('#', '')
            resources['metadata'] = m
            cell.source = ''

        if 'show' in tags:
            cell['metadata']['show_input'] = 'show'

        if 'hide' in tags:
            cell['metadata']['show_input'] = 'hide'

        if cell.cell_type == 'markdown' and re.match('\#\s', cell.source.strip()):
            # Extract the first level 1 heading as the title, if a title isn't already defined
            m = re.match('\s*\#\s*(.*)$', cell.source.strip())
            if m and not self.alt_title:
                if not 'title' in self.front_matter:
                    self.front_matter['title'] = m.groups()[0]
                    cell.source = ''

                self.alt_title = m.groups()[0]

        if cell.cell_type == 'markdown' and 'description' in tags:
            self.front_matter['description'] = cell.source.strip().replace('#', '')
            cell.source = ''

        if cell['source'].startswith('%%metatab'):
            tp = TermParser(TextRowGenerator(re.sub(r'\%\%metatab.*\n', '', cell['source'])),
                            resolver=self.doc.resolver, doc=self.doc)

            self.doc.load_terms(tp)

            # Turn off errors or requests to update Name or Identifier
            cell.outputs = []

        if 'show_input' not in cell['metadata']:
            cell['metadata']['show_input'] = self.front_matter['show_input']

        cell['metadata']['hide_input'] = (cell['metadata']['show_input'] == 'hide')

        return cell, resources



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""

from metapack.cli.core import prt, err
from metapack.package import *
from .core import MetapackCliMemo as _MetapackCliMemo, list_rr
from tabulate import tabulate
import sys
from itertools import islice

downloader = Downloader()

class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)


def stats_args(subparsers):

    parser = subparsers.add_parser(
        'stats',
        help='Create statistical report for a resource'
    )

    parser.add_argument('-n', '--rows', type=int, help='Estimate of number of rows, for sampling')

    parser.add_argument('-s', '--sample', type=int, help='Sample a subset of <SAMPLE> rows')

    parser.add_argument('-H', '--head', type=int, help='Use only the first <HEAD> rows. Can substitute for --rows')

    parser.add_argument('-d', '--descriptive', default=False, action='store_true',
                       help="Calculate descriptive stats; min, max, mean, std, quartiles")

    parser.add_argument('-D', '--distribution', default=False, action='store_true',
                        help="Calculate distribution stats; histogram, skew, kurtosis")

    parser.add_argument('-V', '--values', default=False, action='store_true',
                        help="Display a set of unique values")

    parser.set_defaults(run_command=stats)

    parser.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")


def stats(args):

    m = MetapackCliMemo(args, downloader)

    if m.args.sample and (m.args.head and not m.args.rows):
        m.args.rows = m.args.head

    if bool(m.args.rows) ^ bool(m.args.sample):
        err("Either both or neither of ( --rows or --head) and --sample must be specified")

    if not m.resource:
        prt("ERROR: No resource or reference specified in the URL. Valid resources are:")
        list_rr(m.doc)
        sys.exit(1)

    if m.args.values and (m.args.distribution or m.args.descriptive):
        err("The --values option can be used with neither --descriptive nor --distribution ")

    from tableintuit import Stats

    r = m.doc.resource(m.resource)

    schema = [(c['name'],c['datatype']) for c in r.columns() ]

    if m.args.head:
        source = islice(r.iterdict, m.args.head)
    else:
        source = r.iterdict

    s = Stats( source, schema, n_rows=m.args.rows, sample_size=m.args.sample,
               descriptive=m.args.descriptive,
               distribution=m.args.distribution,
               sample_values=m.args.values)
    s.run()

    if m.args.values:
        for k,v in s.dict.items():
            print("==== {} ".format(k))
            for val, count in v.uvalues.items():
                print("   {} {}".format(count, val))
    else:
        print(s)


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

PACKAGE_PREFIX = '_packages'
MATERIALIZED_DATA_PREFIX='_materialized_data'


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

from metapack import Downloader
from .s3 import metas3
from .metakan import metakan
from .install_file import  install_file
from .install_git import  install_git

downloader = Downloader()

def install(subparsers):

    install_parser = subparsers.add_parser(
        'install',
        help='Install metatap packages',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    install_subparser = install_parser.add_subparsers(help='Install Commands')

    metas3(install_subparser)
    metakan(install_subparser)
    install_file(install_subparser)
    install_git(install_subparser)









# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Create Markdown and HTML of datasets.
"""

import datetime
from rowgenerators import parse_app_url
from markdown import markdown as convert_markdown
from nameparser import HumanName
from pybtex import PybtexEngine
from pybtex.backends.html import Backend as HtmlBackend
from pybtex.style.formatting import toplevel
from pybtex.style.formatting.plain import Style
from pybtex.style.template import (
    words, together, field, sentence, optional_field, href, optional
)
from yaml import safe_dump

from metatab.doc import MetatabDoc

dl_templ = "{}\n:   {}\n\n"

def ns(v):
    """Return empty str if bool false"""
    return str(v) if bool(v) else ''


def linkify(v, description=None, cwd_url=None):


    if not v:
        return None

    u = parse_app_url(v)

    if u.scheme in ('http', 'https', 'mailto'):

        if description is None:
            description = v
        return '[{desc}]({url})'.format(url=v, desc=description)

    elif u.scheme == 'file':

        return '[{desc}]({url})'.format(url=u.path, desc=description)

    else:
        return v


def resource(r, fields=None):
    fields = fields or (
        ('Header', 'header'),
        ('Datatype', 'datatype'),
        ('Description', 'description')
    )

    headers = [f[0] for f in fields]
    keys = [f[1] for f in fields]

    rows = [''.join(["<td>{}</td>".format(e.replace("\n", "<br/>\n")) for e in [c.get(k, '') for k in keys]])
            for c in r.columns()]

    return ("### {name} \n [{url}]({url})\n\n".format(name=r.name, url=r.resolved_url)) + \
           "{}\n".format(ns(r.description)) + \
           "<table class=\"table table-striped\">\n" + \
           ("<tr>{}</tr>".format(''.join("<th>{}</th>".format(e) for e in headers))) + \
           "\n".join("<tr>{}</tr>".format(row) for row in rows) + \
           '</table>\n'


def ckan_resource_markdown(r, fields=None):
    fields = fields or (
        ('Header', 'header'),
        ('Datatype', 'datatype'),
        ('Description', 'description')
    )

    headers = [f[0] for f in fields]
    keys = [f[1] for f in fields]

    def col_line(c):
        return "* **{}** ({}): {}\n".format(
            c.get('header'),
            c.get('datatype'),
            "*{}*".format(c.get('description')) if c.get('description') else '')

    return "### {}. {} Columns. \n\n{}".format(
        ns(r.description),
        len(list(r.columns())),
        ''.join([col_line(c) for c in r.columns()])
    )


def resource_block(doc, fields=None):
    if fields is None:
        fields = [
            ('Header', 'header'),
            ('Datatype', 'datatype'),
            ('Description', 'description')
        ]

        for h in doc['Schema'].args:

            if h.lower() in ('note', 'notes', 'coding'):
                fields.append((h, h.lower()))

    return "".join(resource(r, fields) for r in doc['Resources'].find('Root.Resource'))


def resource_ref(r):
    return dl_templ.format(r.name, "_{}_<br/>{}".format(r.resolved_url, ns(r.description)))


def resource_ref_block(doc):
    return "".join(resource_ref(r) for r in doc['Resources'].find('Root.Resource'))


def contact(r):
    pass


def contacts_block(doc):
    out = ''

    def mklink(url, desc):
        return '[{desc}]({url})'.format(url=url, desc=desc)

    def ctb(url, desc):

        if url and desc:
            return mklink(url, desc)

        elif url:
            return url

        elif desc:
            return desc

        else:
            return ''

    try:

        for t in doc['Contacts']:

            name = t.get_value('name')
            email = "mailto:" + t.get_value('email') if t.get_value('email') else None

            web = t.get_value('url')
            org = t.get_value('organization', web)

            out += (dl_templ.format(t.record_term.title(),
                                    ', '.join(e for e in (ctb(email, name), ctb(web, org)) if e)))

    except KeyError:
        pass

    return out


def documentation_block(doc):
    doc_links = ''

    inline = ''

    notes = []

    try:
        doc['Documentation']
    except KeyError:
        return ''

    try:

        for t in doc['Documentation'].find('Root.IncludeDocumentation'):

            u = parse_app_url(t.value)

            t = doc.package_url.join_target(u).get_resource().get_target()

            with open(t.fspath) as f:
                inline += f.read()

        for t in doc['Documentation'].find('Root.Documentation'):

            title = t.get_value('title')
            desc = t.get_value('description')

            if title and desc:
                dl_templ = "{}\n:   {}\n\n"
            elif title:
                dl_templ = "{}\n\n"
            elif desc:
                title = desc
                dl_templ = "{}\n\n"
            else:
                title = t.value
                dl_templ = "{}\n\n"

            doc_links += (dl_templ.format(linkify(t.resolved_url, title), desc))

        # The doc_img alt text is so we can set a class for CSS to resize the image.
        # img[alt=doc_img] { width: 100 px; }

        for t in doc['Documentation'].find('Root.Image'):
            doc_links += ('[![{}]({} "{}")]({})'
                          .format('doc_img', t.resolved_url, t.get_value('title'),
                                  t.resolved_url))

        for t in doc['Documentation'].find('Root.Note'):
            notes.append(t.value)


    except KeyError:
        raise
        pass

    return inline + \
           (("\n\n## Notes \n\n" + "\n".join('* ' + n for n in notes if n)) if notes else '') + \
           ("\n\n## Documentation Links\n" + doc_links if doc_links else '')




class MetatabStyle(Style):
    # Minnesota Population Center. IPUMS Higher Ed: Version 1.0 [dataset]
    # Minneapolis, MN: University of Minnesota, 2016. http://doi.org/10.18128/D100.V1.0.

    def format_url(self, e):
        return words[
            href[
                field('url', raw=True),
                field('url', raw=True)
            ]
        ]

    def format_accessed(self, e):
        from dateutil.parser import parser

        return words[
            'Accessed',
            field('accessdate', raw=True, apply_func=lambda v: str(parser().parse(v).strftime("%d %b %Y")))
        ]

    def format_dataset(self, e):

        template = toplevel[
            optional[sentence[field('origin')]],
            self.format_btitle(e, 'title'),
            optional[sentence[together['Version', field('version')]]],
            optional[sentence[field('publisher') ]],
            optional[self.format_author_or_editor(e)],
            optional[words[optional_field('month'), field('year')]],
            self.format_web_refs(e),
            self.format_accessed(e)
        ]

        return template.format_data(e)


class MetatabHtmlBackend(HtmlBackend):
    def write_prologue(self):
        pass
        # super().write_prologue()

    def write_epilogue(self):
        pass
        # super().write_epilogue()

    def write_entry(self, key, label, text):
        self.output("<div class='citation'><a name=\"{key}\"><b>[{key}]</b></a> {text} </div>"
                    .format(key=key, text=text))


def make_citation_dict(td):
    """
    Update a citation dictionary by editing the Author field
    :param td: A BixTex format citation dict or a term
    :return:
    """

    from datetime import datetime

    if isinstance(td, dict):
        d = td
        name = d['name']
    else:

        d = td.as_dict()
        d['_term'] = td

        try:
            d['name'] = td.name
        except AttributeError:
            d['name'] = td['name'].value

    if 'author' in d and isinstance(d['author'], str):
        authors = []
        for e in d['author'].split(';'):
            author_d = HumanName(e).as_dict(include_empty=False)
            if 'suffix' in author_d:
                author_d['lineage'] = author_d['suffix']
                del author_d['suffix']
            authors.append(author_d)
        d['author'] = authors

    if not 'type' in d:

        if '_term' in d:
            t = d['_term']

            if t.term_is('Root.Reference') or t.term_is('Root.Resource'):
                d['type'] = 'dataset'

            elif t.term_is('Root.Citation'):
                d['type'] = 'article'
            else:
                d['type'] = 'article'

    if d['type'] == 'dataset':

        if not 'editor' in d:
            d['editor'] = [HumanName('Missing Editor').as_dict(include_empty=False)]

        if not 'accessdate' in d:
            d['accessdate'] = datetime.now().strftime('%Y-%m-%d')

    if not 'author' in d:
        d['author'] = [HumanName('Missing Author').as_dict(include_empty=False)]

    if not 'title' in d:
        d['title'] = d.get('description', '<Missing Title>')

    if not 'journal' in d:
        d['journal'] = '<Missing Journal>'

    if not 'year' in d:
        d['year'] = '<Missing Year>'

    if '_term' in d:
        del d['_term']

    return d


def make_metatab_citation_dict(t):
    """
    Return a dict with BibText key/values for metatab data
    :param t:
    :return:
    """

    try:

        if parse_app_url(t.url).proto == 'metatab':

            try:
                url = parse_app_url(str(t.resolved_url)).resource_url
                doc = t.row_generator.generator.package

            except AttributeError as e:
                return False  # It isn't a resource or reference

            creator = doc.find_first('Root.Creator')
            author_key='author'

            if not creator:
                creator = doc.find_first('Root.Wrangler')
                author_key = 'editor'

            if not creator:
                creator = doc.find_first('Root.Origin')
                author_key = 'editor'

            try:
                origin = doc['Contacts'].get('Root.Origin').get('organization').value
            except AttributeError:
                try:
                    origin = doc.get_value('Root.Origin', doc.get_value('Name.Origin')).title()
                except:
                    origin = None

            d = {
                'type': 'dataset',
                'name': t.name,
                author_key: [HumanName(creator.value).as_dict(include_empty=False)],
                'publisher': creator.properties.get('organization'),
                'origin': origin,
                'journal': '2010 - 2015 American Community Survey',
                'title': doc['Root'].find_first_value('Root.Title') + '; ' + t.name,
                'year': doc.get_value('Root.Year'),
                'accessDate': '{}'.format(datetime.datetime.now().strftime('%Y-%m-%d')),
                'url': url,
                'version': doc.get_value('Root.Version', doc.get_value('Name.Version') ),
            }

            d =  { k:v for k, v in d.items() if v is not None}


            return d


        else:
            return False

    except (AttributeError, KeyError) as e:
        raise
        return False


def _bibliography(doc, terms, converters=[], format='html'):
    """
    Render citations, from a document or a doct of dicts

    If the input is a dict, each key is the name of the citation, and the value is a BibTex
    formatted dict

    :param doc: A MetatabDoc, or a dict of BibTex dicts
    :return:
    """

    output_backend = 'latex' if format == 'latex' else MetatabHtmlBackend

    def mk_cite(v):
        for c in converters:
            r = c(v)

            if r is not False:
                return r

        return make_citation_dict(v)

    if isinstance(doc, MetatabDoc):
        # This doesn't work for LaTex, b/c the formatter adds the prologue and epilogue to eery entry

        d = [mk_cite(t) for t in terms]

        cd = {e['name']: e for e in d}

    else:

        cd = {k: mk_cite(v, i) for i, (k, v) in enumerate(doc.items())}

    # for k, v in cd.items():
    #    print (k, v)

    return PybtexEngine().format_from_string(safe_dump({'entries': cd}),
                                             style=MetatabStyle,
                                             output_backend=output_backend,
                                             bib_format='yaml')


def bibliography(doc, converters=[], format='html'):
    terms = list(doc.get_section('Bibliography', []))

    return _bibliography(doc, terms, converters, format)


def data_sources(doc, converters=[], format='html'):
    terms = list(doc.references()) + list(doc.resources())

    return _bibliography(doc, terms, converters, format)


def identity_block(doc):
    from collections import OrderedDict

    name = doc.find_first('Root.Name')

    d = OrderedDict(
        name=name.value,
        identifier=doc.find_first_value('Root.Identifier'),
    )

    for k, v in name.props.items():
        d[k] = v

    try:
        del d['@value']
    except KeyError:
        pass

    return "".join(dl_templ.format(k, v) for k, v in d.items())


def modtime_str(doc):
    from dateutil.parser import parse

    modtime = doc.find_first_value('Root.Modified') or doc.find_first_value('Root.Issued')

    if modtime:
        try:
            modtime_str = "Last Modified: " + parse(modtime).strftime("%Y-%m-%d %H:%M")
        except ValueError:
            modtime_str = ''

    else:

        modtime_str = ''

    return modtime_str


def markdown(doc):
    """Markdown, specifically for the Notes field in a CKAN dataset"""

    name = doc.find_first('Root.Name')

    text = """

`{name}` {modtime}

_{description}_

{doc_block}

## Contacts

{contacts_block}

""".format(
        title=doc.find_first_value('Root.Title'),
        modtime=modtime_str(doc),
        description=doc.description,
        name=doc.find_first_value('Root.Name'),
        doc_block=documentation_block(doc),
        contacts_block=contacts_block(doc)
    )

    return text


def html(doc):
    extensions = [
        'markdown.extensions.extra',
        'markdown.extensions.admonition'
    ]

    def mdc(text):
        return convert_markdown(text, extensions)

    return """
<html>
<head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <!-- Latest compiled and minified CSS -->
    <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" integrity="sha384-BVYiiSIFeK1dGmJRAkycuHAHRg32OmUcww7on3RYdg4Va+PmSTsz/K68vbdEjh4u" crossorigin="anonymous">

    <!-- Optional theme -->
    <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap-theme.min.css" integrity="sha384-rHyoN1iRsVXV4nD0JutlnGaslCJuC7uwjduW9SVrLvRYooPp2bWYgmgJQIXwl/Sp" crossorigin="anonymous">

    <style>
    .narrow-dl .dl-horizontal dd {{
        margin-left: 100px;
    }}

    .narrow-dl .dl-horizontal dt {{
        width: 80px;
    }}

    </style>
    <title>{title}</title>
</head>
<body>
    <div class="container">
        <div class="row">
            <div class="col-md-12">
                <h1>{title}</h1>
                <p>{modtime}</p>
                <p>{description}</p>
            </div>
        </div>
        <div class="row">
            <div class="col-md-7">
                <h2>Documentation</h2>
                {doc_block}
                <h2>Resources</h2>
                {resource_refs}
            </div>
            <div class="col-md-5 narrow-dl">
                <h2>Identity</h2>
                {identity_block}
                <h2>Contacts</h2>
                {contacts_block}
            </div>
        </div>
        <div class="row">
            <div class="col-md-12">
            {resource_block}
            </div>
        </div>
    </div>

</body>
</html>
    """.format(
        title=doc.find_first_value('Root.Title'),
        modtime=modtime_str(doc),
        description=doc.description,
        identity_block=mdc(identity_block(doc)).replace('<dl>', "<dl class=\"dl-horizontal\">"),
        doc_block=mdc(documentation_block(doc)),
        contacts_block=mdc(contacts_block(doc)).replace('<dl>', "<dl class=\"dl-horizontal\">"),
        resource_refs=mdc(resource_ref_block(doc)).replace('<dl>', "<dl class=\"dl-horizontal\">"),
        resource_block=mdc(resource_block(doc))
    )



#! /usr/bin/env python -u

from string import ascii_uppercase
import sys
import csv
from os import getenv, environ
import json
import select

w = csv.writer(sys.stdout)

w.writerow(['row','type','k','v'])

i=0

for k in sys.argv:
    w.writerow([i,'arg', k, ''])
    i += 1

for k, v in environ.items():
    w.writerow([i,'env',k,v])
    i += 1

for k, v in json.loads(environ['PROPERTIES']).items():
    w.writerow([i, 'prop', k, v])
    i += 1



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN
"""

import sys
from os import getcwd
from os.path import join, basename

from metatab import DEFAULT_METATAB_FILE, resolve_package_metadata_url, MetatabDoc, MetatabError, open_package
from metatab.cli.core import err
from rowgenerators import get_cache, Url
from .core import prt, warn

from metatab.util import slugify
import json
import mimetypes

try:
    import datadotworld as dw
    from datadotworld.client.api import RestApiError
except ImportError:
    err("To run the Metataworld importer, you must first install the datadotworld package. See https://github.com/datadotworld/data.world-py")


def metaworld():
    import argparse
    parser = argparse.ArgumentParser(
        prog='metakan',
        description='Publish packages to Data.World')

    parser.add_argument('-i', '--info', default=False, action='store_true',
                   help="Show package information")


    parser.add_argument('metatabfile', nargs='?', default=DEFAULT_METATAB_FILE, help='Path to a Metatab file')

    class MetapackCliMemo(object):
        def __init__(self, args):
            self.cwd = getcwd()
            self.args = args
            self.cache = get_cache('metapack')

            self.mtfile_arg = args.metatabfile if args.metatabfile else join(self.cwd, DEFAULT_METATAB_FILE)

            self.mtfile_url = Url(self.mtfile_arg)
            self.resource = self.mtfile_url.parts.fragment

            self.package_url, self.mt_file = resolve_package_metadata_url(self.mtfile_url.rebuild_url(False, False))



    m = MetapackCliMemo(parser.parse_args(sys.argv[1:]))

    try:
        doc = MetatabDoc(m.mt_file, cache=m.cache)
    except (IOError, MetatabError) as e:
        err("Failed to open metatab '{}': {}".format(m.mt_file, e))

    if m.args.info:
        package_info(doc)
    else:
        send_to_dw(doc)

    exit(0)


def package_info(doc):

    client = dw.api_client()

    username = 'ericbusboom'

    title = doc.find_first_value("Root.Title")
    key = join(username, slugify(title))

    try:
        ds = client.get_dataset(key)
        prt(json.dumps(ds, indent=4))
    except RestApiError as e:
        err(e)

def get_resource_urls(doc):

    resources = {}

    for dist in doc.find("Root.Distribution"):

        try:
            package_url, metadata_url = resolve_package_metadata_url(dist.value)
        except Exception as e:
            warn("Failed for Distribution {}; {}".format(dist.value, e))
            continue

        u = Url(package_url)

        if u.resource_format == 'zip':
            prt("Skipping ZIP package ", package_url)

        elif u.resource_format == 'xlsx':
            if False:
                resources[basename(package_url)] = package_url
                prt("Adding XLS package ", package_url)
                pass

        elif u.resource_format == 'csv':

            resources[basename(package_url)] = u.signed_resource_url

            prt("Adding CSV package {}".format(basename(package_url)))

            try:
                p = open_package(package_url)
            except (IOError, MetatabError) as e:
                err("Failed to open package '{}' from reference '{}': {}".format(package_url, dist.value, e))

            for r in p.resources():

                mimetype = mimetypes.guess_type(r.resolved_url)[0]

                try:
                    ext = mimetypes.guess_extension(mimetype)[1:]
                except:
                    ext = None

                # '.csv': Data>world currently get the format from the name, not the URL
                resources[r.name+'.csv'] = r.resolved_url
                prt("Adding CSV resource {}".format( r.name))
        else:
            prt('Skipping {}'.format(package_url))


    return resources

def truncate(v, l,suffix=''):


    return v[:(l-len(suffix))] if len(v) > l else v


def send_to_dw(doc):

    client = dw.api_client()

    username = 'ericbusboom'

    title = doc.find_first_value("Root.Title")
    key = username+'/'+slugify(truncate(title,30))

    d = dict(
        title=truncate(title,30),
        description=doc.find_first_value("Root.Description"),
        summary=doc.markdown,
        visibility='OPEN',
        files=get_resource_urls(doc)
    )

    try:

        ds = client.get_dataset(key) # Raise an error if the dataset does not exist

        ds = client.replace_dataset(key, **d)

        ds = client.get_dataset(key)

    except RestApiError:

        ds = client.create_dataset(username, **d)

        ds = client.get_dataset(key)







# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Root program for metapack programs

"""

from pkg_resources import get_distribution, DistributionNotFound, iter_entry_points
import argparse
import logging
import sys

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
   # package is not installed
   pass

from metapack.cli.core import  cli_init

def base_parser():

    parser = argparse.ArgumentParser(
        prog='mp',
        description='Create and manipulate metatab data packages. ')


    parser.add_argument('--exceptions', default=False, action='store_true',
                        help='Show full stack tract for some unhandled exceptions')

    parser.add_argument('--debug', default=False, action='store_true',
                        help='Turn on debug logging ( Basic Config ')


    subparsers = parser.add_subparsers(help='Commands')

    for ep in iter_entry_points(group='mt.subcommands'):

        f = ep.load()
        f(subparsers)

    return parser

def mp():
    from .core import prt, err

    parser = base_parser()

    args = parser.parse_args()

    cli_init(log_level=logging.DEBUG if args.debug else logging.INFO)

    try:
        # Happens when no commands are specified
        args.run_command
    except AttributeError:
        parser.print_help()
        sys.exit(2)


    try:
        args.run_command(args)
    except Exception as e:
        if args.exceptions:
            raise e
        else:
            if e.__cause__:
                err(f"{e.__class__.__name__} {str(e)}\nCaused by: {type(e.__cause__)} {str(e.__cause__)}")
            else:
                err(f"{e.__class__.__name__}  {str(e)}")


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Metapack CLI program for creating new metapack package directories
"""

from metapack.package import *

from .core import MetapackCliMemo as _MetapackCliMemo
import argparse

downloader = Downloader()


class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)

def new_args(subparsers):
    """
    The `mp new` command creates source package directories
    with a proper name, a `.gitignore` file, and optionally, example data,
    entries and code. Typical usage, for creating a new package with most
    of the example options, is ::
    
        mp new -o metatab.org -d tutorial -L -E -T "Quickstart Example Package" 
    
    The :option:`-C` option will set a configuration file, which is a
    Metatab file that with terms that are copied into the `metadata.csv` file
    of the new package. Currently, it copies a limited number of terms,
     including:
    
    - Terms in the Contacts section
    - Root.Space
    - Root.Time
    - Root.Grain
    - Root.Variant
    - Root.Version

    """
    parser = subparsers.add_parser(
        'new',
        help='Create new Metatab packages',
        description=new_args.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.set_defaults(run_command=new_cmd)

    parser.add_argument('-o', '--origin', help="Dataset origin, usually a second-level domain name. Required")
    parser.add_argument('-d', '--dataset', help="Main dataset name. Required", required = True)
    parser.add_argument('-t', '--time', help="Temporal extents, usually a year, ISO8601 time, or interval. ")
    parser.add_argument('-s', '--space', help="Space, geographic extent, such as a name of a state or a Census geoid")
    parser.add_argument('-g', '--grain', help="Grain, the type of entity a row represents")
    parser.add_argument('-v', '--variant', help="Variant, any distinguishing string")
    parser.add_argument('-r', '--revision', help="Version, defaults to 1", default=1)

    parser.add_argument('-T', '--title', help="Set the title")

    parser.add_argument('-L', '--pylib', help="Configure a pylib directory for Python code extensions", action='store_true')

    parser.add_argument('-E', '--example', help="Add examples of resources",
                        action='store_true')

    parser.add_argument('--template', help="Metatab file template, defaults to 'metatab' ", default='metatab')
    parser.add_argument('-C', '--config', help="Path to config file. "
                                               "Defaults to ~/.metapack-defaults.csv or value of METAPACK_DEFAULTS env var."
                                                "Sets defaults for specia root terms and the Contacts section.")


    return parser

def doc_parser():
    from .mp import base_parser

    return new_args(base_parser())

def new_cmd(args):

    from metapack import MetapackDoc
    from metapack.util import make_metatab_file, datetime_now, ensure_dir
    from metapack.cli.core import write_doc, prt, err
    from os.path import exists, join, expanduser
    from metatab import DEFAULT_METATAB_FILE
    from os import getenv

    if args.config:
        config_file = args.config
    elif getenv("METAPACK_CONFIG"):
        config_file = getenv("METAPACK_DEFAULTS")
    elif expanduser('~/.metapack-default.csv'):
        config_file = expanduser('~/.metapack-defaults.csv')
    else:
        config_file = None

    if config_file and exists(config_file):
        prt(f"Using defaults file '{config_file}'")
        config = MetapackDoc(config_file)
    else:
        config = MetapackDoc()

    doc = make_metatab_file(args.template)

    doc['Root']['Created'] = datetime_now()

    origin = args.origin or config.get_value('Root.Origin')

    if not origin:
        err("Must specify a value for origin, either on command line or in defaults file")

    doc['Root'].find_first('Root.Origin').value = origin
    doc['Root'].find_first('Root.Dataset').value = args.dataset
    doc['Root'].find_first('Root.Space').value = args.space or config.get_value('Root.Space')
    doc['Root'].find_first('Root.Time').value = args.time or config.get_value('Root.Time')
    doc['Root'].find_first('Root.Grain').value = args.grain or config.get_value('Root.Grain')
    doc['Root'].find_first('Root.Variant').value = args.variant or config.get_value('Root.Variant')
    doc['Root'].find_first('Root.Version').value = args.revision or config.get_value('Root.Version')

    # Copy contacts in
    if 'Contacts' in config:
        for c in config['Contacts']:
            doc['Contacts'].add_term(c)


    if args.title:
        doc['Root'].find_first('Root.Title').value = args.title.strip()

    nv_name = doc.as_version(None)

    if exists(nv_name):
        err(f"Directory {nv_name} already exists")

    ensure_dir(nv_name)

    doc['Documentation'].new_term('Root.IncludeDocumentation', 'file:README.md', title='README')

    if args.example:
        doc['Resources'].new_term('Root.Datafile',
                                  'http://public.source.civicknowledge.com/example.com/sources/random-names.csv',
                                  name='random_names')

        doc['Documentation'].new_term('Root.Homepage', 'http://metatab.org', title='Metatab Home Page')

    if args.pylib:
        from metapack.support import pylib
        pylib_dir = join(nv_name,'pylib')
        ensure_dir(pylib_dir)
        with open(join(pylib_dir, '__init__.py'), 'w') as f_out, open(pylib.__file__) as f_in:
            f_out.write(f_in.read())

        if args.example:
            doc['Resources'].new_term('Root.Datafile','python:pylib#row_generator', name='row_generator')

    prt(f"Writing to '{nv_name}'")

    write_doc(doc, join(nv_name, DEFAULT_METATAB_FILE))

    with open(join(nv_name,'.gitignore'), 'w') as f:
        f.write("_packages\n")

    if args.title:
        readme = '# {}\n'.format(args.title)
    else:
        readme = '# {}\n'.format(doc.get_value('Root.Name'))

    with open(join(nv_name,'README.md'), 'w') as f:
        f.write(readme)






# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

""" """

from .core import open_package, Downloader
from .filesystem import FileSystemPackageBuilder
from .zip import ZipPackageBuilder
from .s3 import S3PackageBuilder
from .excel import ExcelPackageBuilder
from .csv import CsvPackageBuilder



import unittest
from csv import DictReader

from metapack import MetapackDoc
from metapack.constants import PACKAGE_PREFIX
from metapack import MetapackPackageUrl, MetapackUrl, ResourceError, Downloader
from metapack.cli.core import (make_filesystem_package, make_s3_package, make_excel_package, make_zip_package,
                               make_csv_package,
                               cli_init)
from metapack.test.support import test_data
from metatab.generate import TextRowGenerator
from rowgenerators import get_generator,  parse_app_url
from rowgenerators.exceptions import RowGeneratorError
import platform

downloader = Downloader()


class TestPackages(unittest.TestCase):
    def test_resolve_resource_urls(self):
        """Test how resources are resolved in packages.
            - A name, for excel and CSV packages
            - a path, for ZIP and filesystem packages
            - a web url, for any kind of package
        """
        with open(test_data('packages.csv')) as f:
            for i, l in enumerate(DictReader(f)):


                print(l['url'], l['target_file'])

                u = MetapackPackageUrl(l['url'], downloader=Downloader())

                try:
                    t = u.resolve_url(l['target_file'])
                    self.assertFalse(bool(l['resolve_error']))
                except ResourceError as e:
                    self.assertTrue(bool(l['resolve_error']))
                    continue

                # Testing containment because t can have path in local filesystem, which changes depending on where
                # test is run



                print("   ", t)
                self.assertTrue(l['resolved_url'] in str(t), (i, l['resolved_url'], str(t)))

                try:
                    g = get_generator(t.get_resource().get_target())


                    self.assertEqual(101, len(list(g)))
                    self.assertFalse(bool(l['generate_error']))
                except RowGeneratorError:
                    self.assertTrue(bool(l['generate_error']))
                    continue

    def test_build_package(self):

        try:
            import shapely
            import pyproj
        except ImportError:
            unittest.skip("Pandas not installed")
            return

        cli_init()

        m = MetapackUrl(test_data('packages/example.com/example.com-full-2017-us/metadata.csv'),
                        downloader=downloader)

        package_dir = m.package_url.join_dir(PACKAGE_PREFIX)

        cache = Downloader().cache

        _, fs_url, created = make_filesystem_package(m, package_dir, cache, {}, False)

        print(created)

        return

        fs_url = MetapackUrl('/Volumes/Storage/proj/virt-proj/metapack/metapack/test-data/packages/example.com/' \
                             'example-package/_packages/example.com-example_data_package-2017-us-1/metadata.csv',
                             downloader=downloader)

        # _, url, created =  make_excel_package(fs_url,package_dir,get_cache(), {}, False)

        # _, url, created = make_zip_package(fs_url, package_dir, get_cache(), {}, False)

        # _, url, created = make_csv_package(fs_url, package_dir, get_cache(), {}, False)

        package_dir = parse_app_url('s3://test.library.civicknowledge.com/metatab', downloader=downloader)

        _, url, created = make_s3_package(fs_url, package_dir, cache, {}, False)

        print(url)
        print(created)

    def test_build_simple_package(self):

        cli_init()

        cache = Downloader().cache

        m = MetapackUrl(test_data('packages/example.com/example.com-simple_example-2017-us'), downloader=downloader)

        package_dir = m.package_url.join_dir(PACKAGE_PREFIX)
        package_dir = package_dir

        _, fs_url, created = make_filesystem_package(m, package_dir, cache, {}, False)

        fs_doc = MetapackDoc(fs_url, cache=downloader.cache)

        r = fs_doc.resource('random-names')

        # Excel

        _, url, created = make_excel_package(fs_url, package_dir, cache, {}, False)

        self.assertEquals(['random-names', 'renter_cost', 'unicode-latin1'], [r.name for r in url.doc.resources()])

        self.assertEquals(['random-names', 'renter_cost', 'unicode-latin1'], [r.url for r in url.doc.resources()])

        # ZIP

        _, url, created = make_zip_package(fs_url, package_dir, cache, {}, False)

        self.assertEquals(['random-names', 'renter_cost', 'unicode-latin1'], [r.name for r in url.doc.resources()])

        self.assertEquals(['data/random-names.csv', 'data/renter_cost.csv', 'data/unicode-latin1.csv'],
                          [r.url for r in url.doc.resources()])

        #  CSV

        _, url, created = make_csv_package(fs_url, package_dir, cache, {}, False)

        self.assertEquals(['random-names', 'renter_cost', 'unicode-latin1'], [r.name for r in url.doc.resources()])


        self.assertEquals(
            ['com-simple_example-2017-us-2/data/random-names.csv',
             '.com-simple_example-2017-us-2/data/renter_cost.csv',
             'm-simple_example-2017-us-2/data/unicode-latin1.csv'],
            [str(r.url)[-50:] for r in url.doc.resources()])

    def test_sync_csv_package(self):

        from metapack.package import CsvPackageBuilder

        package_root = MetapackPackageUrl(
            test_data('packages/example.com/example.com-simple_example-2017-us/_packages'), downloader=downloader)

        source_url = 'http://library.metatab.org/example.com-simple_example-2017-us-2/metadata.csv'

        u = MetapackUrl(source_url, downloader=downloader)

        t = u.get_resource().get_target()

        p = CsvPackageBuilder(u, package_root, resource_root=u.dirname().as_type(MetapackPackageUrl))

        csv_url = p.save()

        doc = csv_url.metadata_url.doc

        for r in doc.resources():
            print(r.name, r.url)

            # with open(csv_url.path, mode='rb') as f:
            #    print (f.read())
            #    #urls.append(('csv', s3.write(f.read(), csv_url.target_file, acl)))

    @unittest.skip('References non existen url')
    def test_build_geo_package(self):

        from rowgenerators.valuetype import ShapeValue

        m = MetapackUrl(test_data('packages/sangis.org/sangis.org-census_regions/metadata.csv'), downloader=downloader)

        package_dir = m.package_url.join_dir(PACKAGE_PREFIX)

        _, fs_url, created = make_filesystem_package(m, package_dir, downloader.cache, {}, True)

        print(fs_url)

        doc = MetapackDoc(fs_url)

        r = doc.resource('sra')

        rows = list(r.iterdict)

        self.assertEqual(41, len(rows))

        self.assertIsInstance(rows[1]['geometry'], ShapeValue)

    def test_build_transform_package(self):

        m = MetapackUrl(test_data('packages/example.com/example.com-transforms/metadata.csv'), downloader=downloader)

        package_dir = m.package_url.join_dir(PACKAGE_PREFIX)

        _, fs_url, created = make_filesystem_package(m, package_dir, downloader.cache, {}, False)

        print(fs_url)

    def test_read_geo_packages(self):
        try:
            from publicdata.censusreporter.dataframe import CensusDataFrame
        except ImportError:
            unittest.skip("Public data isn't installed")
            return

        from metapack.jupyter.pandas import MetatabDataFrame
        from metapack.jupyter.pandas import MetatabSeries
        from geopandas.geoseries import GeoSeries
        from rowgenerators.valuetype import ShapeValue

        with open(test_data('line', 'line-oriented-doc.txt')) as f:
            text = f.read()

        doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest\n" + text))

        r = doc.reference('B09020')
        df = r.dataframe()

        self.assertIsInstance(df, CensusDataFrame)

        r = doc.reference('sra_geo')
        df = r.dataframe()

        self.assertIsInstance(df, MetatabDataFrame)

        self.assertIsInstance(df.geometry, MetatabSeries)

        self.assertIsInstance(df.geo.geometry, GeoSeries)

        row = next(r.iterdict)

        self.assertIsInstance(row['geometry'], ShapeValue, type(row['geometry']))


    @unittest.skipIf(platform.system() == 'Windows','Program generators do not work on windows')
    def test_program_resource(self):

        m = MetapackUrl(test_data('packages/example.com/example.com-full-2017-us/metadata.csv'), downloader=downloader)

        doc = MetapackDoc(m)

        r = doc.resource('rowgen')

        self.assertEqual('program+file:scripts/rowgen.py', str(r.url))

        print(r.resolved_url)

        g = r.row_generator

        print(type(g))

        for row in r:
            print(row)

    def test_fixed_resource(self):
        from itertools import islice
        from rowgenerators.generator.fixed import FixedSource

        m = MetapackUrl(test_data('packages/example.com/example.com-full-2017-us/metadata.csv'), downloader=downloader)

        doc = MetapackDoc(m)

        r = doc.resource('simple-fixed')

        self.assertEqual('fixed+http://public.source.civicknowledge.com/example.com/sources/simple-example.txt',
                         str(r.url))
        self.assertEqual('fixed+http://public.source.civicknowledge.com/example.com/sources/simple-example.txt',
                         str(r.resolved_url))

        g = r.row_generator

        print(r.row_processor_table())

        self.assertIsInstance(g, FixedSource)

        rows = list(islice(r, 10))

        print('----')
        for row in rows:
            print(row)

        self.assertEqual('f02d53a3-6bbc-4095-a889-c4dde0ccf5', rows[5][1])

    def test_petl(self):
        from petl import look

        m = MetapackUrl(test_data('packages/example.com/example.com-full-2017-us/metadata.csv'), downloader=downloader)

        doc = MetapackDoc(m)

        r = doc.resource('simple-example')

        t = r.resolved_url.get_resource().get_target()

        p = r.petl()

        print(look(p))

    def test_metapack_resources(self):

        cli_init()

        p = test_data('packages/example.com/example.com-metab_reuse/metadata.csv')

        m = MetapackUrl(p,downloader=downloader)

        print (m.doc.resources())

        print(m.get_resource().get_target().exists())


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""

from os.path import basename, join, dirname, exists
from metapack.exc import MetapackError, ResourceError
from rowgenerators import Url, parse_app_url
from rowgenerators.appurl.web.web import WebUrl
from rowgenerators.appurl.file.file import FileUrl
from rowgenerators.appurl.util import file_ext
from rowgenerators.exceptions import AppUrlError, DownloadError
import json

from metatab import DEFAULT_METATAB_FILE


class _MetapackUrl(object):

    def exists(self):

        if self.inner.proto == 'file':
            return self.inner.exists()
        else:
            # Hack, really ought to implement exeists() for web urls.
            return True

    @property
    def resource_name(self):
        return self.fragment[0]


def is_metapack_url(u):
    return isinstance(u,_MetapackUrl)

class MetapackDocumentUrl(Url, _MetapackUrl):
    def __init__(self, url=None, downloader=None, **kwargs):
        kwargs['proto'] = 'metapack'

        u = Url(url, **kwargs)

        assert downloader

        # If there is no file with an extension in the path, assume that this
        # is a filesystem package, and that the path should have DEFAULT_METATAB_FILE
        if file_ext(basename(u.path)) not in ('zip', 'xlsx', 'csv'):
            u.path = join(u.path, DEFAULT_METATAB_FILE)

        super().__init__(str(u), downloader=downloader, **kwargs)

        self.scheme_extension = 'metapack'

        if basename(self.path) == DEFAULT_METATAB_FILE:
            frag = ''
        elif self.resource_format == 'csv':
            frag = ''
        elif self.resource_format == 'xlsx':
            frag = 'meta'
        elif self.resource_format == 'zip':
            frag = DEFAULT_METATAB_FILE

        self.fragment = [frag, None]

    @classmethod
    def _match(cls, url, **kwargs):
        raise MetapackError("This class should not be contructed through matching")

    @property
    def resource_format(self):

        resource_format = file_ext(basename(self.path))

        assert resource_format, self.path  # Should have either a definite file, or have added one in __init__

        return resource_format

    @property
    def resource_file(self):

        assert basename(self.resource_url)

        return basename(self.resource_url)

    @property
    def target_file(self):
        if self.path.endswith(DEFAULT_METATAB_FILE):
            return DEFAULT_METATAB_FILE
        elif self.resource_format == 'csv':
            return self.resource_file
        elif self.resource_format == 'xlsx':
            return 'meta'
        elif self.resource_format == 'zip':
            return 'metadata.csv'
        else:
            return self.resource_file

    @property
    def target_format(self):
        if self.resource_format == 'csv':
            return 'csv'
        elif self.resource_format == 'xlsx':
            return 'xlsx'
        elif self.resource_format == 'zip':
            return 'csv'
        else:
            return 'csv'

    @property
    def doc(self):
        """Return the metatab document for the URL"""
        from metapack import MetapackDoc
        t = self.get_resource().get_target()
        return MetapackDoc(t, package_url=self.package_url)

    @property
    def generator(self):

        from rowgenerators import get_generator

        t = self.get_resource().get_target().inner

        return get_generator(t)

    def resolve_url(self, resource_name):
        """Return a URL to a local copy of a resource, suitable for get_generator()"""

        if self.target_format == 'csv' and self.target_file != DEFAULT_METATAB_FILE:
            # For CSV packages, need to get the package and open it to get the resoruce URL, becuase
            # they are always absolute web URLs and may not be related to the location of the metadata.
            s = self.get_resource()
            rs = s.doc.resource(resource_name)
            return parse_app_url(rs.url)
        else:
            jt = self.join_target(resource_name)
            rs = jt.get_resource()
            t = rs.get_target()
        return t

    @property
    def metadata_url(self):

        if not basename(self.resource_url):
            return self.clone(path=join(self.path, DEFAULT_METATAB_FILE))
        else:
            return self.clone()

    @property
    def package_url(self):
        """Return the package URL associated with this metadata"""

        if self.resource_file == DEFAULT_METATAB_FILE:
            u = self.inner.clone().clear_fragment()
            u.path = dirname(self.path) + '/'
            u.scheme_extension = 'metapack'
            return MetapackPackageUrl(str(u.clear_fragment()), downloader=self._downloader)
        else:
            return MetapackPackageUrl(str(self.clear_fragment()), downloader=self._downloader)

    def get_resource(self):

        if self.scheme == 'file':
            u = self
        else:
            u = WebUrl(str(self), downloader=self._downloader).get_resource()

        return MetapackDocumentUrl(str(u), downloader=self._downloader)

    def get_target(self):
        return self.inner.get_target()

    def join_target(self, tf):
        if self.target_file == DEFAULT_METATAB_FILE:
            return self.inner.join_dir(tf)
        else:
            return self.inner.join_target(tf)

    @property
    def resource(self):
        """Return the Metapack resource, different from the URL resource returned by get_resource"""

        r = self.doc.resource(self.fragment[0])
        return r


class MetapackPackageUrl(FileUrl, _MetapackUrl):
    """Special version of MetapackUrl for package urls, which never have a fragment"""

    def __init__(self, url=None, downloader=None, **kwargs):
        kwargs['proto'] = 'metapack'

        super().__init__(url, downloader=downloader, **kwargs)
        self.scheme_extension = 'metapack'

        self.fragment = None
        self.query = None

        assert self._downloader

    @property
    def package_url(self):
        return self

    @property
    def metadata_url(self):

        return MetapackDocumentUrl(str(self.clone().clear_fragment()), downloader=self._downloader).metadata_url

    @property
    def doc(self):
        """Return the metatab document for the URL"""
        return self.metadata_url.doc

    def rebuild_fragment(self):
        self.fragment = ''

    def join_resource_name(self, v):
        """Return a MetapackResourceUrl that includes a reference to the resource. Returns a
        MetapackResourceUrl, which will have a a fragment """
        d = self.dict
        d['fragment'] = [v, None]
        return MetapackResourceUrl(downloader=self._downloader, **d)

    def join_resource_path(self, v):
        """Return a regular AppUrl that combines the package with a URL path """
        return self.inner.join(v)

    def join_target(self, tf):
        """Like join(), but returns the inner URL, not a package url class"""
        if self.target_file == DEFAULT_METATAB_FILE:
            return self.inner.join_dir(tf)
        else:
            return self.inner.join_target(tf)

    def resolve_url(self, resource_name):
        """Return a URL to a local copy of a resource, suitable for get_generator()

        For Package URLS, resolution involves generating a URL to a data file from the package URL and the
        value of a resource. The resource value, the url, can be one of:

        - An absolute URL, with a web scheme
        - A relative URL, relative to the package, with a file scheme.

        URLs with non-file schemes are returned. File scheme are assumed to be relative to the package,
        and are resolved according to the type of resource.

        """

        u = parse_app_url(resource_name)

        if u.scheme != 'file':
            t = u

        elif self.target_format == 'csv' and self.target_file != DEFAULT_METATAB_FILE:
            # Thre are two forms for CSV package URLS:
            # - A CSV package, which can only have absolute URLs
            # - A Filesystem package, which can have relative URLs.

            # The complication is that the filsystem package usually has a metadata file named
            # DEFAULT_METATAB_FILE, which can distinguish it from a CSV package, but it's also possible
            # to have a filesystem package with a non standard package name.

            # So, this clause can happed for two cases: A CSV package or a Filesystem package with a nonstandard
            # metadata file name.

            # For CSV packages, need to get the package and open it to get the resource URL, because
            # they are always absolute web URLs and may not be related to the location of the metadata.
            s = self.get_resource()
            rs = s.metadata_url.doc.resource(resource_name)
            if rs is not None:
                t = parse_app_url(rs.url)
            else:
                raise ResourceError("No resource for '{}' in '{}' ".format(resource_name, self))


        else:
            jt = self.join_target(resource_name)
            try:
                rs = jt.get_resource()
            except DownloadError:
                raise ResourceError(
                    "Failed to download resource for '{}' for '{}' in '{}'".format(jt, resource_name, self))
            t = rs.get_target()

        return t

    @property
    def inner(self):
        return super().inner


class MetapackResourceUrl(FileUrl, _MetapackUrl):
    def __init__(self, url=None, downloader=None, base_url=None, **kwargs):

        kwargs['proto'] = 'metapack'

        super().__init__(url, downloader=downloader, **kwargs)
        self.scheme_extension = 'metapack'

        # Expand the path in the same was as the document URL

        # Using .clone() causes recursion
        d = self.dict
        d['fragment'] = None

        md = MetapackDocumentUrl(None, downloader=downloader, **d)
        self.path = md.path

        self.base_url = base_url

    @classmethod
    def _match(cls, url, **kwargs):
        raise MetapackError("This class should not be contructed through matching")

    @property
    def target_file(self):
        from urllib.parse import unquote_plus

        if self.fragment[0]:
            return self.fragment[0]

        return None

    @property
    def doc(self):
        """Return the metatab document for the URL"""
        return self.metadata_url.doc

    @property
    def metadata_url(self):
        assert self._downloader
        return MetapackDocumentUrl(str(self.clone().clear_fragment()), downloader=self._downloader).metadata_url

    @property
    def package_url(self):
        """Return the package URL associated with this metadata"""
        return MetapackDocumentUrl(str(self.clear_fragment()), downloader=self._downloader).package_url

    def get_resource(self):
        if self.scheme == 'file':
            return self
        else:
            u = WebUrl(str(self), downloader=self._downloader)
            r = u.get_resource()
            mru = MetapackResourceUrl(str(r), base_url=self, downloader=self._downloader)
            return mru

    def get_target(self):
        return self.resource.resolved_url.get_resource().get_target()

    @property
    def generator(self):
        # Use fragment b/c it could be target_file, for .zip, or target_segment, for .xlsx
        return self.resource

    @property
    def resource(self):

        r = self.doc.resource(self.fragment[0])
        return r


# Would have made this a function, but it needs to be a class to have the match() method
class MetapackUrl(Url):
    """Implements __new__ to return either a  MetapackResourceUrl or a MetapackDocumentUrl"""

    match_priority = WebUrl.match_priority - 1

    def __new__(cls, url=None, downloader=None, **kwargs):

        assert downloader

        u = Url(url, **kwargs)

        if u.fragment[0]:
            return MetapackResourceUrl(url, downloader, **kwargs)
        else:
            return MetapackDocumentUrl(url, downloader, **kwargs)

    @classmethod
    def _match(cls, url, **kwargs):
        """Return True if this handler can handle the input URL"""
        return url.proto in ('metapack', 'metatab')


class JupyterNotebookUrl(FileUrl):
    """IPYthon Notebook URL"""

    match_priority = FileUrl.match_priority - 10

    def __init__(self, url=None, **kwargs):
        kwargs['proto'] = 'ipynb'
        super().__init__(url, **kwargs)

    @classmethod
    def _match(cls, url, **kwargs):
        return url.resource_format == 'ipynb'

    def get_target(self):
        return self

    def target_dataframe(self):
        if self._target_file:
            return self._target_file

        if self.fragment[0]:
            return self.fragment[0]

        return None


class SearchUrl(Url):
    """Look up a url using a list of callbacks. """

    match_priority = FileUrl.match_priority - 10

    search_callbacks = []

    _search_initialized = False

    def __init__(self, url=None, **kwargs):
        kwargs['scheme'] = 'index'
        super().__init__(url, **kwargs)

    @classmethod
    def _match(cls, url, **kwargs):
        return url.scheme == 'search' or url.scheme == 'index'

    @classmethod
    def register_search(cls, f):
        """Register a function that will lookup the search URL and return a URL if found"""

        cls.search_callbacks.append(f)

    @classmethod
    def initialize(cls):
        if SearchUrl._search_initialized is False:
            from metapack.package.core import Downloader
            try:
                search_func = cls.search_json_indexed_directory(Downloader().cache.getsyspath('/'))
                SearchUrl.register_search(search_func)
            except AppUrlError as e:
                pass

            SearchUrl._search_initialized = True

    @staticmethod
    def search_json_indexed_directory(directory):
        """Return a search function for searching a directory of packages, which has an index.json file
        created by the `mp install file` command.

        This will only search the issued index; it will not return results for the source index
        """

        from metapack.index import SearchIndex, search_index_file

        idx = SearchIndex(search_index_file())

        def _search_function(url):

            packages = idx.search(url, format='issued')

            if not packages:
                return None

            package = packages.pop(0)

            try:
                resource_str = '#' + url.target_file if url.fragment[0] else ''

                return parse_app_url(package['url'] + resource_str, downloader=url.downloader)
            except KeyError as e:
                return None

        return _search_function

    def search(self):
        """Search for a url by returning the value from the first callback that
        returns a non-None value"""

        for cb in SearchUrl.search_callbacks:
            try:
                v = cb(self)
                if v is not None:
                    return v
            except Exception as e:
                raise

    def resolve(self):
        u = self.search()

        if not u:
            raise AppUrlError(f"Search URL failed to resolve reference to {str(self)}")

        u.set_fragment(self.fragment)

        return u

    def get_resource(self):

        return self.resolve().get_resource()

    def get_target(self):
        return self

    def target_dataframe(self):
        if self._target_file:
            return self._target_file

        if self.fragment[0]:
            return self.fragment[0]

        return None



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
CLI program for managing packages
"""

import re

from metapack import MetapackDoc, Downloader
from metapack.cli.core import prt, err, warn, metatab_info, get_lib_module_dict, write_doc, \
    make_excel_package, make_filesystem_package, make_csv_package, make_zip_package, update_name, \
    extract_path_name, MetapackCliMemo, process_schemas
from rowgenerators import parse_app_url
from rowgenerators.exceptions import SourceError
from rowgenerators.util import clean_cache
from tableintuit import RowIntuitError
import argparse

downloader = Downloader()

def build(subparsers):
    """
    Build source packages.

    The mp build program runs all of the resources listed in a Metatab file and
    produces one or more Metapack packages with those resources localized. It
    will always try to produce a Filesystem package, and may optionally produce
    Excel, Zip and CSV packages.

    Typical usage is to be run inside a source package directory with

    .. code-block:: bash

        $ mp build

    To build all of the package types: 
    
    .. code-block:: bash

        $ mp build -fezc

    By default, packages are built with versioned names. The
    :option:`--nonversion-name` option will create file packages with
    non-versioned name, and the :option:`--nonversioned-link` option will
    produce a non-versioned soft link pointing to the versioned file.


    """

    parser = subparsers.add_parser(
        'build',
        help='Build derived packages',
        description=build.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='')


    parser.set_defaults(run_command=run_metapack)

    parser.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv'. "
                          )

    parser.add_argument('-p', '--profile', help="Name of a BOTO or AWS credentails profile", required=False)

    parser.add_argument('-d', '--build-directory', help="Set an alternate build directory, instead of '_packages'",
                            required=False)

    parser.add_argument('-D', '--package-directory', help="Write Zip, Excel and CSV packages to an alternate directory",
                            required=False)

    parser.add_argument('-F', '--force', action='store_true', default=False,
                             help='Force some operations, like updating the name and building packages')

    group = parser.add_mutually_exclusive_group()

    group.add_argument('-n', '--nonversion-name', action='store_true', default=False,
                        help='Write file packages with non-versioned names')

    group.add_argument('-N', '--nonversion-link', action='store_true', default=False,
                        help='Create links with nonversioned names to file packages')

    parser.set_defaults(handler=None)



    ##
    ## Derived Package Group

    derived_group = parser.add_argument_group('Derived Packages', 'Generate other types of packages')

    derived_group.add_argument('-e', '--excel', action='store_true', default=False,
                               help='Create an excel archive from a metatab file')

    derived_group.add_argument('-z', '--zip', action='store_true', default=False,
                               help='Create a zip archive from a metatab file')

    derived_group.add_argument('-f', '--filesystem', action='store_true', default=False,
                               help='Create a filesystem archive from a metatab file')

    derived_group.add_argument('-c', '--csv', action='store_true', default=False,
                               help='Create a CSV archive from a metatab file')

    ##
    ## Administration Group

    admin_group = parser.add_argument_group('Administration', 'Information and administration')

    admin_group.add_argument('--clean-cache', default=False, action='store_true',
                             help="Clean the download cache")

    admin_group.add_argument('-C', '--clean', default=False, action='store_true',
                             help="For some operations, like updating schemas, clear the section of existing terms first")


def run_metapack(args):
    from rowgenerators.rowpipe.exceptions import TooManyCastingErrors

    m = MetapackCliMemo(args, downloader)


    if m.args.profile:
        from metatab.s3 import set_s3_profile
        set_s3_profile(m.args.profile)

    try:
        for handler in ( metatab_derived_handler, metatab_admin_handler):
            handler(m)
    except TooManyCastingErrors as e:
        prt('Casting Errors:')
        for error in e.errors:
            prt(error)
        if m.args.exceptions:
            raise e
        else:
            err(e)
    except Exception as e:
        if m.args.exceptions:
            raise e
        else:
            err(e)

    clean_cache(m.cache)


def metatab_derived_handler(m):
    """Create local Zip, Excel and Filesystem packages

    :param m:
    :param skip_if_exists:
    :return:
    """
    from metapack.exc import PackageError
    from metapack.util import get_materialized_data_cache
    from shutil import rmtree

    create_list = []
    url = None

    doc = MetapackDoc(m.mt_file)

    env = get_lib_module_dict(doc)

    package_dir = m.package_root

    if m.args.package_directory:
        # If this is set, the FS package will be built to m.package_root, but the
        # file packages will be built to package_dir
        package_dir = parse_app_url(m.args.package_directory)

    update_name(m.mt_file, fail_on_missing=False, report_unchanged=False)

    process_schemas(m.mt_file, cache=m.cache, clean=m.args.clean, report_found=False)

    nv_name = m.args.nonversion_name
    nv_link = m.args.nonversion_link

    # Remove any data that may have been cached , for instance, from Jupyter notebooks
    rmtree(get_materialized_data_cache(doc), ignore_errors=True)

    try:

        # Always create a filesystem package before ZIP or Excel, so we can use it as a source for
        # data for the other packages. This means that Transform processes and programs only need
        # to be run once.

        # all_build_opts = [m.args.filesystem, m.args.excel, m.args.zip, m.args.csv]

        _, url, created = make_filesystem_package(m.mt_file, m.package_root, m.cache, env, m.args.force, False, nv_link)
        create_list.append(('fs', url, created))

        m.mt_file = url

        env = {}  # Don't need it anymore, since no more programs will be run.

        if m.args.excel is not False:
            _, url, created = make_excel_package(m.mt_file, package_dir, m.cache, env, m.args.force, nv_name, nv_link)
            create_list.append(('xlsx', url, created))

        if m.args.zip is not False:
            _, url, created = make_zip_package(m.mt_file, package_dir, m.cache, env, m.args.force, nv_name, nv_link)
            create_list.append(('zip', url, created))

        if m.args.csv is not False:
            _, url, created = make_csv_package(m.mt_file, package_dir, m.cache, env, m.args.force, nv_name, nv_link)
            create_list.append(('csv', url, created))

    except PackageError as e:
        err("Failed to generate package: {}".format(e))

    return create_list


def metatab_admin_handler(m):

    if m.args.clean_cache:
        clean_cache('metapack')


def classify_url(url):
    ss = parse_app_url(url)

    if ss.target_format in DATA_FORMATS:
        term_name = 'DataFile'
    elif ss.target_format in DOC_FORMATS:
        term_name = 'Documentation'
    else:
        term_name = 'Resource'

    return term_name


def add_resource(mt_file, ref, cache):
    """Add a resources entry, downloading the intuiting the file, replacing entries with
    the same reference"""

    if isinstance(mt_file, MetapackDoc):
        doc = mt_file
    else:
        doc = MetapackDoc(mt_file)

    if not 'Resources' in doc:
        doc.new_section('Resources')

    doc['Resources'].args = [e for e in set(doc['Resources'].args + ['Name', 'StartLine', 'HeaderLines', 'Encoding']) if
                             e]

    seen_names = set()

    u = parse_app_url(ref)

    # The web and file URLs don't list the same.

    if u.proto == 'file':
        entries = u.list()
    else:
        entries = [ssu for su in u.list() for ssu in su.list()]

    print(type(u), u)

    for e in entries:
        add_single_resource(doc, e, cache=cache, seen_names=seen_names)

    write_doc(doc, mt_file)


def add_single_resource(doc, ref, cache, seen_names):
    from metatab.util import slugify

    t = doc.find_first('Root.Datafile', value=ref)

    if t:
        prt("Datafile exists for '{}', deleting".format(ref))
        doc.remove_term(t)
    else:
        prt("Adding {}".format(ref))

    term_name = classify_url(ref)

    path, name = extract_path_name(ref)

    # If the name already exists, try to create a new one.
    # 20 attempts ought to be enough.
    if name in seen_names:
        base_name = re.sub(r'-?\d+$', '', name)

        for i in range(1, 20):
            name = "{}-{}".format(base_name, i)
            if name not in seen_names:
                break

    seen_names.add(name)

    encoding = start_line = None
    header_lines = []



    if not name:
        from hashlib import sha1
        name = sha1(slugify(path).encode('ascii')).hexdigest()[:12]

        # xlrd gets grouchy if the name doesn't start with a char
        try:
            int(name[0])
            name = 'a' + name[1:]
        except:
            pass

    return doc['Resources'].new_term(term_name, ref, name=name,
                                     startline=start_line,
                                     headerlines=','.join(str(e) for e in header_lines),
                                     encoding=encoding)


def run_row_intuit(path, cache):
    from tableintuit import RowIntuiter
    from itertools import islice
    from rowgenerators import TextEncodingError, get_generator

    for encoding in ('ascii', 'utf8', 'latin1'):
        try:

            u = parse_app_url(path)
            u.encoding = encoding

            rows = list(islice(get_generator(url=str(u), cache=cache, ), 5000))
            return encoding, RowIntuiter().run(list(rows))
        except (TextEncodingError, UnicodeEncodeError) as e:
            pass

    raise RowIntuitError('Failed to convert with any encoding')


DATA_FORMATS = ('xls', 'xlsx', 'tsv', 'csv')
DOC_FORMATS = ('pdf', 'doc', 'docx', 'html')



# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

from __future__ import print_function

from os.path import exists
from uuid import uuid4

from metatab.cli.core import prt
from metatab.cli.metapack import classify_url
from metatab.doc import MetatabDoc
from metatab.util import make_metatab_file


# Change the row cache name


def init_metatab(mt_file, alt_mt_file):
    mt_file = alt_mt_file if alt_mt_file else mt_file

    prt("Initializing '{}'".format(mt_file))

    if not exists(mt_file):
        doc = make_metatab_file()

        doc['Root']['Identifier'] = str(uuid4())

        doc.write_csv(mt_file)
    else:
        prt("Doing nothing; file '{}' already exists".format(mt_file))


def scrape_page(mt_file, url):
    from .util import scrape_urls_from_web_page

    doc = MetatabDoc(mt_file)

    doc['resources'].new_term('DownloadPage', url)

    d = scrape_urls_from_web_page(url)

    for k, v in d['sources'].items():
        doc['Resources'].new_term('DataFile', v['url'], description=v.get('description'))

    for k, v in d['external_documentation'].items():
        term_name = classify_url(v['url'])
        doc['Documentation'].new_term(term_name, v['url'], description=v.get('description'))

    doc.write_csv(mt_file)


def rewrite_resource_format(mt_file):
    doc = MetatabDoc(mt_file)

    if 'schema' in doc:
        table_schemas = {t.value: t.as_dict() for t in doc['schema']}
        del doc['schema']

        for resource in doc['resources']:
            s = resource.new_child('Schema', '')
            for column in table_schemas.get(resource.find_value('name'), {})['column']:
                c = s.new_child('column', column['name'])

                for k, v in column.items():
                    if k != 'name':
                        c.new_child(k, v)

    doc.write_csv(mt_file)





from __future__ import print_function

import unittest

from rowgenerators import parse_app_url
from metapack import open_package, MetapackDoc, Downloader
from metapack.constants import PACKAGE_PREFIX
from metatab.util import flatten
from metatab.generate import TextRowGenerator
from itertools import islice
from rowgenerators import get_generator
from metapack.test.support import test_data, get_cache


import logging
from metapack.cli.core import cli_init
import os

logger = logging.getLogger('user')
logger_err = logging.getLogger('cli-errors')
debug_logger = logging.getLogger('debug')

downloader = Downloader()

class TestIPython(unittest.TestCase):
    def compare_dict(self, a, b):

        fa = set('{}={}'.format(k, v) for k, v in flatten(a));
        fb = set('{}={}'.format(k, v) for k, v in flatten(b));

        # The declare lines move around a lot, and rarely indicate an error
        fa = {e for e in fa if not e.startswith('declare=')}
        fb = {e for e in fb if not e.startswith('declare=')}

        errors = len(fa - fb) + len(fb - fa)

        if errors:
            print("=== ERRORS ===")

        if len(fa - fb):
            print("In b but not a")
            for e in sorted(fa - fb):
                print('    ', e)

        if len(fb - fa):
            print("In a but not b")
            for e in sorted(fb - fa):
                print('    ', e)

        self.assertEqual(0, errors)


    def test_html(self):

        p = open_package(test_data('packages/example.com/example.com-full-2017-us/metadata.csv'))

        self.assertTrue(len(p._repr_html_()) > 6400, len(p._repr_html_()) )

        print ( list(e.name for e in p.find('Root.Resource')))

        r = p.find_first('Root.Resource', name='random-names')

        self.assertTrue(len(r._repr_html_()) > 400, len(r._repr_html_()))

    def test_dataframe(self):

        try:
            import pandas
        except ImportError:
            unittest.skip("Pandas not installed")
            return

        p = open_package(test_data('packages/example.com/example.com-full-2017-us/metadata.csv'))

        r = p.resource('random-names')

        df = r.dataframe()

        print (df.describe())

    @unittest.skip("Requires Publicdata module")
    def test_line_doc(self):

        from os.path import splitext, basename, join
        import sys


        with open(test_data('line','line-oriented-doc.txt')) as f:
            text = f.read()

        doc = MetapackDoc(TextRowGenerator("Declare: metatab-latest\n" + text))

        #process_schemas(doc)

        r = doc.reference('tracts')

        self.assertEqual(628, len(list(r)))

        tracts = r.dataframe()

        self.assertEqual(-73427, tracts.lon.sum().astype(int))

        tracts = r.read_csv()

        self.assertEqual(-73427, tracts.lon.sum().astype(int))

        r.dataframe()

        # Test loading a Python Library from a package.

        ref = doc.reference('incv')

        self.assertIsNotNone(ref)

        ref_resource = parse_app_url(ref.url).inner.clear_fragment().get_resource()

        # The path has to be a Metatab ZIP archive, and the root directory must be the same as
        # the name of the path

        pkg_name, _ = splitext(basename(ref_resource.path))

        lib_path = ref_resource.join(pkg_name).path

        if lib_path not in sys.path:
            sys.path.insert(0, lib_path)



    def test_notebook_url(self):

        try:
            import pandas
            import jupyter_client
        except ImportError:
            unittest.skip("Missing pandas or jupyter client")
            return

        from metapack.appurl import JupyterNotebookUrl
        from metapack.jupyter.exec import execute_notebook
        from os.path import exists

        u = parse_app_url(test_data('notebooks','GenerateDataTest.ipynb'))

        self.assertIsInstance(u, JupyterNotebookUrl)

        nb = execute_notebook(u.path, '/tmp/nbtest', ['dfa','dfb'], True)

        self.assertTrue(exists('/tmp/nbtest/dfa.csv'))
        self.assertTrue(exists('/tmp/nbtest/dfb.csv'))

        g = get_generator(parse_app_url('/tmp/nbtest/dfa.csv'))

        print(list(g))

    @unittest.skipIf("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true", "Skipping this test on Travis CI.")
    def test_build_notebook_package(self):

        try:
            import pandas
        except ImportError:
            unittest.skip("Pandas not installed")
            return

        from metapack import MetapackUrl, MetapackDocumentUrl
        from metapack.cli.core import make_filesystem_package, process_schemas

        m = MetapackDocumentUrl(test_data('packages/example.com/example.com-notebook/metadata.csv'), downloader=downloader)

        #process_schemas(m)

        doc = MetapackDoc(m)

        r = doc.resource('basic_a')

        self.assertEqual(2501, len(list(r)))

        package_dir = m.package_url.join_dir(PACKAGE_PREFIX)

        _, fs_url, created = make_filesystem_package(m, package_dir, get_cache(), {}, False, False, False)

        print(fs_url)

    @unittest.skip("Broken")
    def test_nbconvert(self):

        from collections import namedtuple
        from metapack.jupyter.convert import convert_documentation

        cli_init()

        M = namedtuple('M', 'mt_file')

        fn = test_data('notebooks/ConversionTest.ipynb')

        m = M(mt_file=parse_app_url(fn))

        convert_documentation(m.mt_file.path)

    @unittest.skip('Pandoc & latex wackiness')
    def test_nbconvert_package(self):

        try:
            import pandoc
        except (ImportError, FileNotFoundError):
            unittest.skip("Pandoc is not installed")
            return

        from collections import namedtuple
        from metapack.jupyter.convert import convert_notebook

        cli_init()

        M = namedtuple('M', 'mt_file mtfile_arg init_stage2')

        fn = test_data('notebooks/ConversionTest.ipynb')

        m = M(mt_file=parse_app_url(fn), mtfile_arg=parse_app_url(fn),
              init_stage2=lambda x,y: None)

        convert_notebook(m.mt_file.path)

    @unittest.skip("Broken")
    def x_test_pandas(self):

        package_dir = '/Volumes/Storage/proj/virt-proj/metatab3/metatab-packages/civicknowledge.com/immigration-vs-gdp'

        doc = open_package(package_dir)

        r = doc.first_resource(name='country_gdp')

        rows = list(r)

        print(len(rows))

        df = r.dataframe()

        print(df.head())

    @unittest.skip("Broken")
    def x_test_metatab_line(self):
        from metatab.generate import TextRowGenerator
        from metatab.cli.core import process_schemas

        cli_init()

        doc = MetatabDoc(TextRowGenerator(test_data('simple-text.txt'), 'simple-text.txt'))

        process_schemas(doc)

        r = doc.resource('resource')

        for c in r.columns():
            print(c)

    @unittest.skip("Broken")
    def x_test_ipy(self):
        from rowgenerators import SourceSpec, Url, RowGenerator, get_cache

        urls = (
            'ipynb+file:foobar.ipynb',
            'ipynb+http://example.com/foobar.ipynb',
            'ipynb:foobar.ipynb'

        )

        for url in urls:
            u = Url(url)
            print(u, u.path, u.resource_url)

            s = SourceSpec(url)
            print(s, s.proto, s.scheme, s.resource_url, s.target_file, s.target_format)
            self.assertIn(s.scheme, ('file', 'http'))
            self.assertEquals('ipynb', s.proto)

        gen = RowGenerator(cache=get_cache(),
                           url='ipynb:scripts/Py3Notebook.ipynb#lst',
                           working_dir=test_data(),
                           generator_args={'mult': lambda x: x * 3})

        rows = gen.generator.execute()

        print(len(rows))


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

""" """

from os.path import join, dirname, exists, abspath, islink
from os import unlink, symlink
from metapack import PackageError
from metapack.util import datetime_now
from metapack.package.core import PackageBuilder
from rowgenerators import parse_app_url


class CsvPackageBuilder(PackageBuilder):
    """"""

    type_code = 'csv'
    type_suffix = '.csv'

    def __init__(self, source_ref=None, package_root=None,  resource_root=None, callback=None, env=None):

        super().__init__(source_ref, package_root,  callback, env)
        from metapack import MetapackPackageUrl

        self.package_path, self.cache_path = self.make_package_path(self.package_root, self.package_name)

        if self.package_root.proto == 'file':
            self.package_root.ensure_dir()

        if resource_root is not None:
            self.resource_root = resource_root
        else:
            self.resource_root = source_ref.dirname().as_type(MetapackPackageUrl)

        assert isinstance(self.resource_root, MetapackPackageUrl), (type(self.resource_root), self.resource_root)

        self._last_write_path = None

    @classmethod
    def make_package_path(cls, package_root, package_name):

        # self.package_path, self.cache_path = self.make_package_path(self.package_name, self.package_root)

        cache_path = package_name + cls.type_suffix

        package_path = package_root.join(cache_path)

        return package_path, cache_path


    def _load_resource(self, source_r, abs_path=False):
        """The CSV package has no reseources, so we just need to resolve the URLs to them. Usually, the
            CSV package is built from a file system ackage on a publically acessible server. """

        r = self.doc.resource(source_r.name)

        r.url = self.resource_root.join(r.url).inner #r.resolved_url


    def _relink_documentation(self):

        for doc in self.doc['Documentation'].find(['Root.Documentation', 'Root.Image']):
            doc.url =  doc.resolved_url

    def save(self, path=None):
        from metapack import MetapackPackageUrl

        # HACK ...
        if not self.doc.ref:
            self.doc._ref = self.package_path  # Really should not do this but ...

        self.check_is_ready()

        self.load_declares()

        self.doc.cleanse()

        self._load_resources()

        self._relink_documentation()

        self._clean_doc()

        if path is None:
            if self.package_path.inner.proto == 'file':
                path = self.package_path.path
            else:
                raise PackageError("Can't write doc to path: '{}'".format(path))

        self.doc['Root'].get_or_new_term('Root.Issued').value = datetime_now()

        self._last_write_path = path

        self.doc.write_csv(path)

        return parse_app_url(abspath(path)).as_type(MetapackPackageUrl)









class NotebootUrl(Url):
    """IPYthon Notebook URL"""

    def __init__(self, url, **kwargs):
        kwargs['proto'] = 'ipynb'

        super(NotebootUrl, self).__init__(url, **kwargs)

        raise NotImplementedError("Is this used?") # Added 20171201


    @classmethod
    def match(cls, url, **kwargs):
        return extract_proto(url) == 'ipynb'

    def _extract_parts(self, url, kwargs):
        parts = self.url_parts(url, **kwargs)

        self.url = reparse_url(url, assume_localhost=True,
                               scheme=parts.scheme if parts.scheme != 'ipynb' else 'file',
                               scheme_extension='ipynb')

        self.parts = self.url_parts(self.url, **kwargs)

    @property
    def path(self):
        return self.parts.path

    def _process_resource_url(self):
        self.resource_url = unparse_url_dict(self.parts.__dict__,
                                             scheme=self.parts.scheme if self.parts.scheme != 'ipynb' else 'file',
                                             scheme_extension=False,
                                             fragment=False).strip('/')

        self.resource_file = basename(self.resource_url)

        if not self.resource_format:
            self.resource_format = file_ext(self.resource_file)

    def _process_fragment(self):
        self.target_segment = self.parts.fragment

    def _process_target_file(self):
        super(NotebootUrl, self)._process_target_file()

        assert self.target_format == 'ipynb', self.target_format




# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""

from rowgenerators import Source
from metapack.jupyter.exec import execute_notebook
from os.path import join, exists
from csv import reader

from rowgenerators.exceptions import AppUrlError
from metapack.util import get_materialized_data_cache

class JupyterNotebookSource(Source):
    """Generate rows from an IPython Notebook.

     This generator will execute a Jupyter notebook. Before execution, it adds an "%mt_materalize"
     magic to the notebook, which will cause the target dataframe to be written to a temporary file, and
     the temporary file is yielded to the caller. Not most efficient, but it fits the model.
     """

    def __init__(self, ref, cache=None, working_dir=None, env=None, doc=None, **kwargs):
        super().__init__(ref, cache, working_dir, **kwargs)

        self.env = env
        self.doc = doc

    def start(self):
        pass

    def finish(self):
        pass


    def __iter__(self):

        from os.path import basename

        self.start()

        dr_name = get_materialized_data_cache(self.doc)

        path = join(dr_name,  self.ref.target_file + ".csv")

        # The first python notebook in the resources will get executed, cache the datafiles,
        # then the remaining ones will just use the caches data.
        if not exists(path):

            # The execute_notebook() function will add a cell with the '%mt_materialize' magic,
            # with a path that will case the file to be written to the same location as
            # path, below.

            if not self.ref.target_dataframe():
                raise AppUrlError('Url did not specify a dataframe; use the "#" fragment ')

            nb = execute_notebook(self.ref.path, dr_name, [self.ref.target_dataframe()], True, self.env)

        with open(path) as f:
            yield from reader(f)

        self.finish()


class PandasDataframeSource(Source):
    """Iterates a pandas dataframe  """


    def __init__(self, url, df, cache, working_dir=None, **kwargs):
        super().__init__(url, cache, working_dir, **kwargs)

        self._df = df

    def __iter__(self):

        import numpy as np

        self.start()

        df = self._df

        if len(df.index.names) == 1 and df.index.names[0] is None and df.index.dtype != np.dtype('O'):
            # For an unnamed, single index, assume that it is just a row number
            # and we don't really need it

            yield list(df.columns)

            for index, row in df.iterrows():
                yield list(row)

        else:

            # Otherwise, either there are more than

            index_names = [n if n else "index{}".format(i) for i, n in enumerate(df.index.names)]

            yield index_names + list(df.columns)

            if len(df.index.names) == 1:
                idx_list = lambda x: [x]
            else:
                idx_list = lambda x: list(x)

            for index, row in df.iterrows():
                yield idx_list(index) + list(row)


        self.finish()


def copy_reference(resource, doc, env, *args, **kwargs):
    """A row-generating function that yields from a reference. This permits an upstream package to be
    copied and modified by this package, while being formall referenced as a dependency

    The function will generate rows from a reference that has the same name as the resource term
    """

    yield from doc.reference(resource.name)





# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN

The program uses the Root.Distributions in the source package to locate packages to link into a CKAN record.

"""


from metapack.package import *
from metapack.index import SearchIndex, search_index_file
from metapack.exc import MetatabFileNotFound
from rowgenerators import parse_app_url
from rowgenerators.appurl.file import FileUrl
from rowgenerators.appurl.web import S3Url
from rowgenerators.exceptions import RowGeneratorError
from .core import MetapackCliMemo as _MetapackCliMemo, new_search_index
from .core import err, prt, debug_logger
from metapack.constants import PACKAGE_PREFIX

from tabulate import tabulate

downloader = Downloader()


class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)
        self


def index_args(subparsers):
    parser = subparsers.add_parser(
        'index',
        help='Index packages for searching. '
    )

    parser.set_defaults(run_command=index)

    parser.add_argument('-l', '--list', default=False, action='store_true',
                        help="List the packages that would be indexed")

    parser.add_argument('-c', '--clear', default=False, action='store_true',
                        help="Clear the index")

    parser.add_argument('-C', '--config', default=False, action='store_true',
                        help="Show the location of the index file")

    parser.add_argument('-d', '--directory', default=downloader.cache.getsyspath('/'),
                        help="Directory where index will be stored")

    parser.add_argument('-D', '--dump',
                        help="Dump a Metatab formatted version of the index")

    parser.add_argument('-L', '--load',
                        help="Load in a Metatab formatted index file ")

    parser.add_argument('metatab_url', nargs='?', default='./',
                        help='URL to a metatab package or container for packages')

def walk_packages(args, u):
    from os import walk
    from os.path import islink, join, isdir

    seen = set()

    if not isdir(u.path):
        try:
            yield open_package(u.path)
        except (RowGeneratorError, MetatabFileNotFound) as e:
            pass

        return

    for root, dirs, files in walk(u.path):

        try:
            p = open_package(root)
            # This was a package, so only recurse if it is a source package and has a _packages dir

            if PACKAGE_PREFIX in dirs:
                del dirs[:]
                dirs.append(PACKAGE_PREFIX)
            else:
                del dirs[:]

            if str(p.ref) not in seen:
                yield p
            seen.add(str(p.ref))
            continue

        except (RowGeneratorError, MetatabFileNotFound) as e:
            # directory is not a package, carry on
            pass

        for f in files:
            if not islink(join(root,f)):
                try:
                    p = open_package(join(root,f))
                    if str(p.ref) not in seen:
                        yield p
                    seen.add(str(p.ref))
                except (RowGeneratorError, MetatabFileNotFound) as e:
                    # directory is not a package, carry on
                    pass


def index(args):

    idx = SearchIndex(search_index_file())

    u = parse_app_url(args.metatab_url)

    if args.list:

        pkg_list = []
        for p in walk_packages(args, u):
            pkg_list.append((p.name, p.ref))

        prt(tabulate(pkg_list, headers='name Url'.split()))
    elif args.config:
        import shlex
        prt(idx.path)

    elif args.clear:
        idx.clear()
        prt('Cleared the index')
    elif args.dump:
        dump_index(args, idx)
    elif args.load:
        load_index(args, idx)
    elif isinstance(u, FileUrl):
        entries = []
        for p in walk_packages(args, u):
            idx.add_package(p)
            entries.append(p.name)

        idx.write()
        prt("Indexed ", len(entries), 'entries')

    elif isinstance(u, S3Url):
        # S3 package collections are flat, so we don't have to walk recursively.
        # However, we are only going to take the S3 packages, because they have the distribution
        # information for the rest of the packages, so we can get info about Excel and Zip packages
        # without opening them.
        entries = []
        import re

        for e in u.list():
            if e.target_format == 'csv':

                p = open_package(e)

                for d in p.find('Root.Distribution'):
                    u = parse_app_url(d.value)

                    version_m = re.search('-([^-]+)$', p.name)

                    if u.target_format in ('xlsx','zip','csv'):
                        idx.add_entry(p.identifier, p.name, p.nonver_name, version_m.group(1), u.target_format,
                                      'metapack+'+str(u))



                        entries.append(p.name)
        idx.write()

        prt("Indexed ",len(entries), 'entries')

    else:
        err(f"Can only index File and S3 urls, not {type(u)}")

def dump_index(args, idx):

    import csv
    import sys

    from metatab import MetatabDoc

    doc = MetatabDoc()

    pack_section = doc.new_section('Packages', ['Identifier', 'Name', 'Nvname', 'Version', 'Format'])

    r = doc['Root']
    r.new_term('Root.Title', 'Package Index')

    for p in idx.list():

        pack_section.new_term('Package',
                   p['url'],
                   identifier=p['ident'],
                   name=p['name'],
                   nvname=p['nvname'],
                   version=p['version'],
                   format=p['format'])

    doc.write_csv(args.dump)

def load_index(args, idx):
    import json

    from metatab import MetatabDoc

    doc = MetatabDoc(args.load)

    entries = set()

    for t in doc['Packages']:
        idx.add_entry(t.identifier, t.name, t.nvname, t.version, t.format, t.value)
        entries.add(t.value)

    prt("Loaded {} packages".format(len(entries)))
    idx.write()


# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""

from metatab.exc import MetatabError

class MetapackError(MetatabError):
    pass

class MetatabFileNotFound(MetatabError):
    pass

class PackageError(MetapackError):
    pass

class ResourceError(MetapackError):
    pass



